
//============================================================================
//
//                              Program/Module
//                                   from
//                C++ package for verified numerical integration
//                  of singular and strong singular functions
//                      Copyright (c) 2000 Stefan Wedner
//
//         This program/module is free software for non-commercial use.
//
//
//============================================================================
//----------------------------------------------------------------------------
//
// File:     peano.tpl
// See also: peano.h, peano.C
//     
// ---------------------------------------------------------------------------


// #define power opower

// template<class T> T opower(const T&  x, int n){
//  T result(1);  
//  if (n > 0){   
//   if (n % 2){
//     for(int i=1; i<=n; i++) result *= x;
//     return result;
//   }
//   T tmp = sqr(x);
//   n /= 2;
//   for(int i=1; i<=n; i++) result *= tmp;
//   return result;
//  }
//  if (n == 0) return result; 
//  if (n == -1) return 1/x;
//  if ( n == -2) return 1 / sqr(x);
//  cout << " power error " << endl; 
// }


// ----------------------------------------------------------------------------
// class peano_weight member functions
// ----------------------------------------------------------------------------


// --------------------- constructor ------------------------------------------

template<> peano_weight<interval>::peano_weight(string filename, int precision,
                                                real eps)
:peano_base(filename, false)
// init nodes and weights for T = interval
{
  epsilon = eps;
 
  weight       = new interval[number_nodes+1];
  node         = new interval[number_nodes+1];

  for(int i=1; i<=number_nodes;i++){
     node[i]   = peano_base::node[i];
     weight[i] = peano_base::weight[i];
  }
}


template<class T> peano_weight<T>::peano_weight(string filename, int precision,
                                                real eps):peano_base()
{
  // init nodes and weights for T = l_interval
  int oldprecision = stagprec;

  epsilon   = eps; 
  stagprec  = precision;
  symmetric = false;

  ifstream external_data(filename.data());

  if(! external_data) cout << " error! : file " << filename 
                           << " not available! " << endl;

  external_data >> number_nodes;
  peano_base::node   =  new interval[number_nodes+1];
  peano_base::weight =  new interval[number_nodes+1];
  node   = new T[number_nodes+1];
  weight = new T[number_nodes+1];

  readdata(external_data, number_nodes, node, weight, peano_base::node, 
                                                      peano_base::weight);
  external_data.close();
  stagprec = oldprecision;
}


// --------------------- destructor -------------------------------------------

template<class T> peano_weight<T>::~peano_weight(){
  for(int i=1; i< order_kernel; i++) delete[] s_over_nu[i];
  delete[] s_over_nu;  
  delete[] node;
  delete[] weight;
}


// --------------------- member functions kernel_alpha ------------------------

template<class T> T peano_weight<T>::kernel_alpha(real r, int s){
  // compute int_t^1 (t-x)^s / sqrt(x) dx ( s! gamma_{s+1} )
 T x(r), help;
 int k, pm = 2;
 
 if (r == 0){
   help = 2 * s + 1;
   return 2.0 / help;
 }
 if ( r == 1){ help = 0; return help;}
 i_dot = 0;
 for(k=0; k<=s; k++){
  help = 2 * k + 1;
  accumulate(i_dot, s_over_nu[s][k], pm / help);
  pm *= -1;
 }
 help = i_dot;
 help *= power(x, s) * sqrt(x);
 if (s%2) i_dot = help;
 else     i_dot = -help;
 
 for(k=0; k<=s; k++)
  if ( (s - k) % 2)
   accumulate(i_dot, -2.0 * s_over_nu[s][k]/ (2*k + 1) , power(x, s-k));
  else
   accumulate(i_dot,  2.0 * s_over_nu[s][k]/ (2*k + 1) , power(x, s-k));
 help = i_dot;
 return help;
}


// --------------------- member functions kernel ------------------------------

template<class T> interval peano_weight<T>::kernel(int k){
  // compute (s-1)! * K_s([x_k])
  real x_inf = Inf(peano_base::node[k]), x_sup = Sup(peano_base::node[k]);
  T low, up;
  T diff;
  int s = order_kernel -1;
  
  low = kernel_alpha(x_inf, s);
  up  = kernel_alpha(x_sup, s);

  i_dot = T(Inf(up), Sup(low));

  for(int  i= start_index; i<=number_nodes; i++ ){
    diff = truncated_diff(node[i], node[k]);
    diff = power(diff, order_kernel - 1);
    accumulate(i_dot, -weight[i], diff);
  }
  interval result; rnd(i_dot,result);
  return result;
}


template<class T> interval peano_weight<T>::kernel(interval x){
  // compute (m-1)! * K_m([x]) ; m = order_kernel
  T low,  up;
  interval diff;

  int s = order_kernel -1;
 
  if (start_index <= (number_nodes + 1) / 2){
    low   = x;
    i_dot = power(low, s) * sqrt(low) * gamma;
      //  second part of kernel:  
    for(int  i= 1; i< start_index; i++ )
      accumulate(i_dot, -weight[i], power(x - node[i], order_kernel - 1) );
    rnd(i_dot, diff);
    if (order_kernel % 2) diff *= -1;
    return diff;
  }

  low   = kernel_alpha(Inf(x), s);
  up    = kernel_alpha(Sup(x), s);
  i_dot = T(Inf(up), Sup(low));

  //  second part of kernel:  
  for(int  i= start_index; i<=number_nodes; i++ )
    accumulate(i_dot, -weight[i], power(node[i] - x, order_kernel - 1) );
  rnd(i_dot, diff);
  return diff;
}


// --------------------- member function derivation1 --------------------------

template<class T> interval peano_weight<T>::derivation1(interval x){
 // compute (m-1)! * K'_m([x]) ; m = order_kernel
  T low,  up, diffT;
  interval diff;
  int s = order_kernel - 2;

  if (start_index <= (number_nodes + 1) / 2 || s < 0){

    low = x;
    i_dot = power(low, s) * sqrt(low) * gamma * (order_kernel - 0.5);

    for(int  i= 1; i < start_index;  i++ ){
      if ( order_kernel == 2) diff = 1;
      else if ( order_kernel == 1) diff = 0;
        else diff = (order_kernel - 1) * power( x - node[i], order_kernel - 2);
      accumulate(i_dot, -weight[i], diff);
    }
    rnd(i_dot, diff);
    if (order_kernel % 2) diff *= -1;
    return diff;
  }

  low = kernel_alpha(Inf(x), s);
  up  =  kernel_alpha(Sup(x),s);

  i_dot = -(s + 1) * T(Inf(up), Sup(low));

  for(int  i= start_index; i<=number_nodes; i++ ){
    if ( order_kernel == 2) diffT = 1;
    else if (order_kernel == 1) diffT = 0; 
     else diffT  = (order_kernel - 1) * power(node[i] - x, order_kernel - 2);
    accumulate(i_dot, weight[i], diffT);
  }

  rnd(i_dot, diff);
  return diff;
}


// --------------------- member function derivation2 --------------------------

template<class T> interval peano_weight<T>::derivation2(interval x){
 // compute (m-1)! * K"_m([x]) ; m = order_kernel
  T low,  up, diffT;
  interval diff;
  int s = order_kernel - 3;

  if (start_index <= (number_nodes + 1) / 2 || s < 0){

    low = x;
    i_dot = power(low, s) * sqrt(low) * gamma * (order_kernel - 0.5) 
                                            * (order_kernel - 1.5);
    for(int  i= 1; i < start_index; i++ ){
      if( order_kernel < 3 ) diff = 0;
      else  if ( order_kernel == 3)  diff = 2;
            else diff = (order_kernel-2) * (order_kernel-1)*
                        power(x - node[i], s);
      accumulate(i_dot, -weight[i], diff);
    }

    rnd(i_dot, diff );
    if (order_kernel % 2) diff *= -1;
    return diff;
  }

  low = kernel_alpha(Inf(x),s);
  up  = kernel_alpha(Sup(x),s);

  i_dot = (s + 2) * (s + 1) * T(Inf(up), Sup(low));

  for(int  i= start_index; i<=number_nodes; i++ ){
    if( order_kernel < 3 ) diffT = 0;
    else  if ( order_kernel == 3)  diffT = 2;
          else diffT = (order_kernel-2) * (order_kernel-1)*
                        power(node[i]-x, order_kernel-3);
     accumulate(i_dot, -weight[i], diffT);
  }

  rnd(i_dot, diff );
  return diff;
}


// --------------------- member function integrate_stamm ----------------------


template<class T> void peano_weight<T>::integrate_stamm(real from, real to){
  // compute int_from^to K_m([x]) ; m = order_kernel
  T low,  up;
  interval  diff; 
 
  int s = order_kernel;

  low = kernel_alpha(from, s);
  up  = kernel_alpha(to, s);

  i_dot = low - up;
  for( int i=start_index; i<=number_nodes; i++ ) {  
    diff =truncated_diff( power(node[i]-from, order_kernel), 
                           power(node[i]-to, order_kernel));
    accumulate( i_dot, -weight[i], diff);
  }

  rnd( i_dot, diff);

  if( Inf(diff)>= 0 )
    positive_part += i_dot;
  else if( Sup(diff)<=0 ) 
    negative_part += i_dot;
  else {
    positive_part +=  interval(0,Sup(diff));
    negative_part +=  interval(Inf(diff),0); 
  }
}



// --------------------- member function init_data ----------------------------


template<class T> void peano_weight<T>::init_data(){

  int i,j;

  s_over_nu = new T*[order_kernel+1];

  for(i = 0; i <= order_kernel; i++){
      s_over_nu[i] = new T[order_kernel+1];
  }

  for(i = 0; i <= order_kernel; i++){
      s_over_nu[0][i] = 0;
      s_over_nu[i][0] = 1;
  }
  for(i = 1; i <= order_kernel; i++){
      s_over_nu[1][i] = 0; 
      s_over_nu[i][1] = i;
  }
   
  for(i = 2; i <=  order_kernel; i++){
    for(j = 2; j <=  i ; j++)  
        s_over_nu[i][j] = s_over_nu[i-1][j-1] + s_over_nu[i-1][j] ; 
    for(j = i+1; j <=  order_kernel ; j++)  s_over_nu[i][j] = 0; 
  } 

  gamma  =  2;
  gamma /= (2 * order_kernel - 1); 
  for(i = 1; i < order_kernel; i++){
    gamma *= i * 2.0;
    gamma /= (2*i -1);
  }
}



// --------------------- high test function------------------------------------


template<class T> T peano_weight<T>::high(){
 
  T m;
 
  m = 4 * number_nodes + 1;
  m = 2 / m;
  i_dot = m;
  for(int i=1; i<=number_nodes; i++){
    accumulate(i_dot,power(node[i], 2 * number_nodes), -weight[i]);
  } 
  m = i_dot;
  return m;
}




// ----------------------------------------------------------------------------
// class peano_cauchy member functions
// ----------------------------------------------------------------------------



// --------------------- constructor ------------------------------------------


template<> peano_cauchy<interval>::peano_cauchy(string filename, int precision,
                                                real eps) :peano_base(filename)
{
  epsilon   = eps;
  symmetric = true;
  weight        = new interval[number_nodes+1];
  node          = new interval[number_nodes+1];
  int i,m = number_nodes;
  if (m % 2){
    m /= 2;
    for(i=1; i<= m; i++){
       node[i]    = peano_base::node[i];
       weight[i]  = peano_base::weight[i] = peano_base::weight[i] / node[i];
    }
    weight[m+1] = peano_base::weight[m+1] = 0;
    for(i=m+2; i<=number_nodes; i++){
       node[i]    = peano_base::node[i];
       weight[i]  = peano_base::weight[i] = peano_base::weight[i] / node[i];
    } 
  }
  else{
    for(i=1; i<= number_nodes; i++){  
       node[i]    = peano_base::node[i];
       weight[i]  = peano_base::weight[i] = peano_base::weight[i] / node[i];
       } 
    }
}


template<class T> peano_cauchy<T>::peano_cauchy(string filename, int precision,
                                                real eps):peano_base()
{
  epsilon = eps;
  int i, m, old_precision = stagprec;
  
  stagprec  = precision;
  symmetric = true;
   
  ifstream external_data(filename.data());

  if(!external_data) cout << " error! : file " << filename 
                          << " not available! " << endl;

  external_data >> number_nodes;
  m = number_nodes;

  // allocation heap
  peano_base::node   =  new interval[number_nodes+1];
  peano_base::weight =  new interval[number_nodes+1];

  node   = new T[number_nodes+1];
  weight = new T[number_nodes+1];

  readdata_symm(external_data, number_nodes, node, weight, peano_base::node, 
                                                         peano_base::weight);
  if (m % 2){
     m /= 2;
    for(i=1; i<= m; i++){
      peano_base::weight[i] /= peano_base::node[i]; 
      weight[i]  /=  node[i];
    }
;
    weight[m+1] = peano_base::weight[m+1] = 0;
    
    for(i=m+2; i<=number_nodes; i++){
       peano_base::weight[i] /= peano_base::node[i];
       weight[i]  /= node[i];
    }
  }
  else{
    for(i=1; i<= number_nodes; i++){
       weight[i]             /= node[i];
       peano_base::weight[i] /= peano_base::node[i]; 
    }
  }
  stagprec = old_precision;
}


// --------------------- destructor -------------------------------------------

template<class T> peano_cauchy<T>::~peano_cauchy(){

  delete[]  sum_nu ;
  for(int i=1; i< order_kernel; i++) delete[] s_over_nu[i];
  delete[]  s_over_nu;  
  delete[] node;
  delete[] weight;
}



// --------------------- member function kernel_beta --------------------------

template<class T> T peano_cauchy<T>::kernel_beta(real x, int s){

  if (x <= 0) return T(0);
  T help(x);  
  if (s % 2)
    return power(help, s) * (ln(help) - sum_nu[s]);
  else 
    return power(help, s) * (sum_nu[s] - ln(help));
}


// --------------------- member functions kernel ------------------------------

template<class T> interval peano_cauchy<T>::kernel(int k){

 // modified: low, up must be greater than 0 (fi_lib!)

  real x_inf = Inf(peano_base::node[k]), x_sup = Sup(peano_base::node[k]);
  T low(x_inf),  up(x_sup), diffT;
  interval diff;
  int s = order_kernel -1;
  
  i_dot = 0;

  for(int nu = 1; nu < s; nu++){
    if ( (s - nu) % 2) 
      accumulate(i_dot, -power(low, s - nu), s_over_nu[s][nu] / nu);
    else 
      accumulate(i_dot,  power(low, s - nu), s_over_nu[s][nu] / nu);
  }
  accumulate(i_dot,  s_over_nu[s][s] / s, T(1));
  accumulate(i_dot, 1.0, kernel_beta(x_inf, s));
  low = i_dot;
 
  i_dot = 0;

  for(int nu = 1; nu < s; nu++){ 
   if ( (s - nu) % 2) 
    accumulate(i_dot, - power(up, s - nu), s_over_nu[s][nu] / nu);
   else
    accumulate(i_dot,   power(up, s - nu), s_over_nu[s][nu] / nu);
  }  
  accumulate(i_dot, 1.0 , s_over_nu[s][s] / s);
  accumulate(i_dot, 1.0, kernel_beta(x_sup, s));
  up = i_dot;

  i_dot = T(Inf(up), Sup(low));

  for(int  i= start_index; i<=number_nodes; i++ ){
    diffT = truncated_diff(node[i], node[k]);
    diffT = power(diffT, order_kernel - 1);
    accumulate(i_dot, -weight[i], diffT);
  }
  rnd(i_dot, diff);
  return diff;
}



template<class T> interval peano_cauchy<T>::kernel(interval x){
  T low( Inf(x)),  up(Sup(x));
  interval diff;
  int s = order_kernel -1;
 
  //  alpha(low)
  i_dot = 0;
  for(int nu = 1; nu < s; nu++){
    if ( (s - nu)  % 2)
      accumulate(i_dot, -power(low, s - nu), s_over_nu[s][nu] / nu);
    else 
      accumulate(i_dot,  power(low, s - nu), s_over_nu[s][nu] / nu); 
  }
  accumulate(i_dot, 1.0, 1.0 / T(s,s));
 
  //  beta(low)
  accumulate(i_dot, 1.0, kernel_beta(Inf(x), s));

  low = i_dot;

  //  alpha(up)
  i_dot = 0;
  for(int nu = 1; nu < s; nu++){
    if ( (s - nu) % 2) 
      accumulate(i_dot, -power(up, s - nu), s_over_nu[s][nu] / nu);
    else 
      accumulate(i_dot,  power(up, s - nu), s_over_nu[s][nu] / nu);
  }
  accumulate(i_dot, 1.0 , 1.0 / T(s,s) );

  //  beta(up)
  accumulate(i_dot, 1.0, kernel_beta(Sup(x), s));
  up = i_dot;
  
  //  gamma(x)
  i_dot = T(Inf(up), Sup(low));

  //  gamma(x)  
  for(int  i= start_index; i<=number_nodes; i++ )
    accumulate(i_dot, -weight[i], power(node[i] - x, order_kernel - 1) );
  rnd(i_dot, diff);
  return diff;
}


// --------------------- member function derivation1 --------------------------

template<class T> interval peano_cauchy<T>::derivation1(interval x){
  T low(Inf(x)),  up(Sup(x)), diffT;
  interval diff;
  int s = order_kernel - 2;

  // alpha_s(low)
  i_dot = 0;

  if (s!=0){ 
    for(int nu = 1; nu < s; nu++){ 
      if ( (s- nu) % 2)
        accumulate(i_dot, -power(low, s - nu), s_over_nu[s][nu] / nu);
      else
        accumulate(i_dot, power(low, s - nu), s_over_nu[s][nu] / nu);
    }
    accumulate(i_dot, 1.0 , s_over_nu[s][s] / s);
  }
   // beta_s(low)
  accumulate(i_dot, 1.0 , kernel_beta(Inf(x), s));
  low = i_dot;
 
  // alpha_s(up)
  i_dot = 0;

  if (s!=0){ 
   for(int nu = 1; nu < s; nu++){
     if ( (s - nu) % 2) 
       accumulate(i_dot, -power(up, s - nu), s_over_nu[s][nu] / nu);
     else
       accumulate(i_dot,  power(up, s - nu), s_over_nu[s][nu] / nu);
   }
   accumulate(i_dot, 1.0, 1.0 / T(s,s));
  }
  // beta_s(up)
  accumulate(i_dot, 1.0, kernel_beta(Sup(x),s));
  up = i_dot;

  i_dot = -(s + 1) * T(Inf(up), Sup(low));

  for(int  i= start_index; i<=number_nodes; i++ ){
    if ( order_kernel == 2) diffT = 1;
    else diffT  = (order_kernel - 1) * power( node[i] - x, order_kernel - 2);
    accumulate(i_dot, weight[i], diffT); // - removed
  }
  rnd(i_dot, diff);
  return diff;
}


// --------------------- member function derivation2 --------------------------

template<class T> interval peano_cauchy<T>::derivation2(interval x){
  T low( Inf(x) ),  up(Sup(x)), diffT;
  interval diff;
  int s = order_kernel - 3;

  i_dot = 0;

  if (s == 0){ 
    low = -2.0 * ln(low); 
    up =  -2.0 * ln(up); 
    i_dot = T(Inf(up), Sup(low));
  }
  else if (s < 0){
    low =  -1.0 / low;
    up  =  -1.0 / up;
    i_dot = T(Inf(up), Sup(low));
  }
  else{
    for(int nu = 1; nu < s; nu++){
      if ( (s - nu) % 2) 
        accumulate(i_dot, -power(low, s - nu), s_over_nu[s][nu] / nu);
      else 
        accumulate(i_dot,  power(low, s - nu), s_over_nu[s][nu] / nu);
    }
    accumulate(i_dot, 1.0 , s_over_nu[s][s] / s);
  
    accumulate(i_dot, 1.0, kernel_beta(Inf(x), s));
    low = i_dot;

    i_dot = 0;

    for(int nu = 1; nu < s; nu++){
      if ( (s - nu) % 2) 
        accumulate(i_dot, -power(up, s - nu), s_over_nu[s][nu] / nu);
      else
        accumulate(i_dot,  power(up, s - nu), s_over_nu[s][nu] / nu);
    } 
    accumulate(i_dot, 1.0 , s_over_nu[s][s] / s);
    accumulate(i_dot, 1.0, kernel_beta(Sup(x),s));
    up = i_dot;
    i_dot = (s + 2) * (s + 1) * T(Inf(up), Sup(low));
  }

  for(int  i= start_index; i<=number_nodes; i++ ){
    if( order_kernel < 3 ) diffT = 0;
    else  if ( order_kernel == 3)  diffT = 2;
          else diffT = (order_kernel-2) * (order_kernel-1)*
                       power(node[i]-x, order_kernel-3);
     accumulate(i_dot, -weight[i], diffT);
  }
  rnd(i_dot, diff );
  return diff;
}


// --------------------- member function integrate_stamm ----------------------

template<class T> void peano_cauchy<T>::integrate_stamm(real from, real to){
 
  T low(from),  up(to); 
  interval diff; 
  int s = order_kernel;


  i_dot = 0;
  for(int nu = 1; nu < s; nu++){
    if ( (s - nu) % 2) 
      accumulate(i_dot, -power(low, s - nu), s_over_nu[s][nu] / nu);
    else
      accumulate(i_dot,  power(low, s - nu), s_over_nu[s][nu] / nu);
  }
  accumulate(i_dot, 1.0 , s_over_nu[s][s] / s);
  accumulate(i_dot, 1.0, kernel_beta(from, s));
  low = i_dot;
 
  i_dot = 0;
  for(int nu = 1; nu < s; nu++){
    if( (s - nu) % 2)  
      accumulate(i_dot, -power(up, s - nu), s_over_nu[s][nu] / nu);
    else
      accumulate(i_dot,  power(up, s - nu), s_over_nu[s][nu] / nu);
  }
  accumulate(i_dot, 1.0 , s_over_nu[s][s] / s);
  accumulate(i_dot, 1.0, kernel_beta(to, s));
  up = i_dot;

  i_dot = low - up;

  for( int i=start_index; i<=number_nodes; i++ ) {  
    diff =truncated_diff( power(node[i]-from, order_kernel), 
                     power(node[i]-to, order_kernel));
    accumulate( i_dot, -weight[i], diff);
  }
  rnd( i_dot, diff);
  if( Inf(diff)>= 0 )
    positive_part += i_dot;
  else if( Sup(diff)<=0 ) negative_part += i_dot;
  else {
    positive_part +=  interval(0,Sup(diff));
    negative_part +=  interval(Inf(diff),0); 
  }
 //  cout << " stamm " << interval(from, to)  << endl <<  " stamm+ " <<  diff  << endl; 
}


// --------------------- member function use_symmetrie ------------------------

template<class T> void peano_cauchy<T>::use_symmetrie(){
 if (!odd(order_kernel)){
    i_dot = positive_part;
    positive_part -= negative_part;
    negative_part -= i_dot;
  }
  else {
      positive_part += positive_part; 
      negative_part += negative_part;
  }
}



// --------------------- member function init_data ----------------------------

template<class T> void peano_cauchy<T>::init_data(){

  int i,j;
  sum_nu = new T[order_kernel+1];
  s_over_nu = new T*[order_kernel+1];

  sum_nu[0] = 0;

  for(i = 1; i <= order_kernel; i++){
      s_over_nu[i] = new T[order_kernel+1];
      sum_nu[i] = sum_nu[i-1] + 1 / T(i,i);
  }
  for(i = 1; i <= order_kernel; i++){
      s_over_nu[1][i] = 0; 
      s_over_nu[i][1] = i;
  }
   
  for(i = 2; i <=  order_kernel; i++){
    for(j = 2; j <=  i ; j++)  
        s_over_nu[i][j] = s_over_nu[i-1][j-1] + s_over_nu[i-1][j] ; 
    for(j = i+1; j <=  order_kernel ; j++)  s_over_nu[i][j] = 0; 
  } 
}
