//============================================================================
//
//                              Program/Module
//                               from CLAVIS
//                C++ library for verified numerical integration
//                           of singular functions
//                     Copyright (c) 2000 Stefan Wedner
//
//          This program/module is free software for non-commercial use. 
//
// 
//============================================================================
//----------------------------------------------------------------------------
//
// File:    f_ari.C
// Purpose: taylorarithmetic (one dimensional and two dimensional)
//
//============================================================================
// 
//


#include"f_ari.h"
#include<iostream>


// ===========================================================================
// ---------------------------------------------------------------------------
// class operand
// ---------------------------------------------------------------------------
// ===========================================================================

bool operand::detail = false;
bool operand::done   = false;
int  operand::dim_var1;
int  operand::dim_var2;
bool operand::triangular = false;
int  operand::s_lb;
int  operand::s_ub;
int  operand::t_lb;
int  operand::t_ub;   
idotprecision operand::i_accu;
operand* operand::var1;
operand* operand::var2;
operand* operand::named[2 * max_named];
int operand::index_named = 0;


#ifdef  MEMORY_CHECK
 int erz = 0, dest = 0;
int testdest(){ return erz - dest;} 
#endif
// ----------------------------------------------------------------------------
// constructors and destructor
// ----------------------------------------------------------------------------

operand::operand(operand& ri){
  o_type = ri.o_type;
  left   = ri.left;
  right  = ri.right;
  value  = ri.value;
  dep_s  = ri.dep_s;
  dep_t  = ri.dep_t;
  status = false;
  del    = false;
  tk     = 0;
  if(ri.del) ri.delete_operand();
  else throw;   // not a valid assignment, sorry not implemented
}

operand::operand(operand_type otype, operand& Left, operand& Right){
  o_type  = otype;
  left    = &Left;
  right   = &Right;
  dep_s   = (Left.dep_s || Right.dep_s) ;
  dep_t   = (Left.dep_t || Right.dep_t);
  status  = false;
  value   = 0;
  del     = true;
  tk      = 0;
#ifdef  MEMORY_CHECK
  erz++; 
  cout << " erzeugt: " <<o_type2string(otype) << this << endl;
#endif
}


operand::operand(const operand* op, operand* Left, operand* Right){
  o_type   = op->o_type;
  left     = Left;
  right    = Right;
  dep_s    = op->dep_s;
  dep_t    = op->dep_t;
  status   = false;
  exponent = op->exponent;
  value    = op->value;
  del      = op->del;
  tk       = 0;
#ifdef  MEMORY_CHECK
  erz++; 
  cout << " erzeugt: " <<o_type2string( o_type) << this << endl;
#endif
}


operand::operand(operand_type otype, operand& Left){
 o_type  = otype;
 left    = &Left;
 right   = 0;
 dep_s   = Left.dep_s;
 dep_t   = Left.dep_t;
 status  = false;
 value   = 0;
 del     = true;
 tk      = 0;
#ifdef  MEMORY_CHECK
 erz++; 
 cout << " erzeugt: " <<o_type2string( o_type) << this << endl;
#endif
}



operand::operand(operand_type otype, bool deps, bool dept){
 o_type  = otype;
 left    = 0;
 right   = 0;
 dep_s   = deps;
 dep_t   = dept;
 status  = false;
 value   = 0;
 del     = true;
 tk      = 0;
#ifdef  MEMORY_CHECK
 erz++; 
 cout << " erzeugt: " <<o_type2string( o_type) << this << endl;
#endif
}


operand::operand(operand_type t_var){
  // generate operand for variable
  o_type  = t_var;
  left    = 0;
  right   = 0;
  del     = false; 
  if (t_var == Var_1){
    dep_t   = false;
    dep_s   = true;
  }
  else{
    dep_t   = true;
    dep_s   = false;
  }
  tk      = 0;
  status  = false;
  value   = 0;
#ifdef  MEMORY_CHECK
  erz++; 
  cout << " erzeugtbenannt: " <<o_type2string(o_type) << this << endl;
#endif
}


operand::operand(const real& constant){
  o_type  = Const;
  left    = 0;
  right   = 0;
  value   = constant;
  del     = true;
  dep_t   = false;
  dep_s   = false;
  status  = false;
  tk      = 0;
#ifdef  MEMORY_CHECK
  erz++;
  cout << " erzeugt: " <<o_type2string( o_type) << this << endl;
#endif
}


operand::operand(const interval& constant){
  o_type  = Const;
  left    = 0;
  right   = 0;
  value   = constant;
  dep_t   = false;
  dep_s   = false;
  status  = false;
  del     = true;
  tk      = 0;
#ifdef  MEMORY_CHECK
  erz++; 
  cout << " erzeugt: " << o_type2string( o_type) << this << endl; 
#endif
}


operand::~operand(){
#ifdef  MEMORY_CHECK
  dest++;
  cout << " Destruktor operand: " << o_type2string(o_type) << this << endl;
#endif
  if(left && left->del){ delete left; left = 0;}
  if(right && right->del){ delete right; right = 0;}

}



// ----------------------------------------------------------------------------
// internal functions set_detail 
// ----------------------------------------------------------------------------


void operand::set_detail(bool Detail){
  detail = Detail;
}



// ----------------------------------------------------------------------------
// internal functions set_value
// ----------------------------------------------------------------------------

void operand::set_value(){

  if (status == done) return;
  switch (o_type){

  case Uminus: left->set_value();
               value = -left->value; break;
  case Plus  : left->set_value();
               right->set_value();
               value = left->value + right->value; break;
  case Minus : left->set_value();
               right->set_value();
               value = left->value - right->value; break;
  case Mult  : left->set_value();
               right->set_value();
               value = left->value * right->value; break;
  case Div   : left->set_value();
               right->set_value();
               value = left->value / right->value; break;
  case Pow   : left->set_value();
               value = pow(left->value, right->value);
               right->status = done; break;
  case Power : left->set_value(); 
               value = power(left->value, right->exponent);
               right->status = done; break;
  case Exp   : left->set_value();
               value = exp(left->value); break;
  case Ln    : left->set_value();
               value = ln(left->value); break;
  case Sqr   : left->set_value();
               value = sqr(left->value); break;
  case Sqrt  : left->set_value();
               value = sqrt(left->value); break;
  case Sin   : left->set_value();
               value = sin(left->value);
               right->status = done; break;
  case Cos   : left->set_value();
               value = cos(left->value);
               right->status = done; break;
  case Tan   : left->set_value();
               value = tan(left->value);
               right->status = done; right->left->status =done;
               right->left->right->status = done; break;
  case Cot   : left->set_value();
               value = cot(left->value);
               right->status = done; right->left->status =done;
               right->left->right->status = done; break;              
  case Asin  : left->set_value();
               value = asin(left->value);
               right->status = done; right->left->status = done;
               right->left->left->status = done; 
               right->left->right->status = done; break;
  case Acos  : left->set_value();
               value = acos(left->value);
               right->status = done; right->left->status = done;
               right->left->left->status = done; 
               right->left->right->status = done; break;
  case Atan  : left->set_value();
               value = atan(left->value);
               right->status = done; right->left->status = done;
               right->right->status = done; break;
  case Acot  : left->set_value();
               value = acot(left->value);
               right->status = done; right->left->status = done;
               right->right->status = done; break;
  case Sinh  : left->set_value();
               value = sinh(left->value);
               right->status = done; break;
  case Cosh  : left->set_value();
               value = cosh(left->value);
               right->status = done; break;
  case Tanh  : left->set_value();
               value = tanh(left->value);
               right->status = done; right->left->status =done;
               right->left->right->status = done; break;
  case Coth  : left->set_value();
               value = coth(left->value);
               right->status = done; right->left->status =done;
               right->left->right->status = done; break;
  case Asinh : left->set_value();
               value = asinh(left->value);
               right->status = done; right->left->status =done;
               right->left->left->status = done;
               right->left->right->status = done; break;
  case Acosh : left->set_value();
               value = acosh(left->value);
               right->status = done; right->left->status =done;
               right->left->left->status = done;
               right->left->right->status = done; break;
  case Atanh : left->set_value();
               value = atanh(left->value);
               right->status = done; right->left->status = done;
               right->right->status = done; break;
  case Acoth : left->set_value();
               value = acoth(left->value);
               right->status = done; right->left->status = done;
               right->right->status = done; break;

  default    : break;
  }
  status = done;
}



// ----------------------------------------------------------------------------
// internal functions trigo1, trigo2 (internal function used by diff)
// ----------------------------------------------------------------------------

void operand::trigo1(i_function stf){

   int i,j,k,l;
   interval tmp, tmp2;

   if (!dep_t){ // stf(u(s))
     i = s_lb;
     if(!i){ tk[0] = stf(left->tk[0]); i++; }
     for(; i<= s_ub; i++){
        tmp = 0;
        for(k = 1; k < i; k++) 
          tmp += k * tk[k] * right->tk[i-k];             
        tk[i] = (left->tk[i] - tmp / i ) / right->tk[0];  
     }
     return;   
   }
   if (!dep_s){ // stf(u(t)
     j = t_lb;
     if(!j){ tk[0] = stf(left->tk[0]); j++; }
     for(; j<= t_ub; j++){
        tmp = 0;
        for(l = 1; l < j; l++) 
          tmp += l * tk[l] * right->tk[j-l];             
        tk[j] = (left->tk[j] - tmp / j ) / right->tk[0];  
     }
     return;   
   }
       // stf(u(s,t))
   TK[0][0] = stf(left->TK[0][0]);
   for(j = 1; j <= t_ub; j++){
      tmp = 0;
      for(l = 1; l < j; l++) 
        tmp += l * TK[0][l] * right->TK[0][j-l];             
      TK[0][j] = (left->TK[0][j] - tmp / j ) / right->TK[0][0];
   }
   for(i = 1; i<=s_ub; i++) 
     for(j = 0; j<=t_ub; j++){
       tmp2 = 0;
       for(k = 1; k < i; k++){
         tmp = 0;
         for(l = 0; l <= j; l++)
           tmp += right->TK[i-k][j-l] * TK[k][l]; 
           tmp2 += k * tmp;                          
       }
       tmp2 /= i;
       tmp = 0;
       for(l = 1; l <= j; l++)
          tmp += right->TK[0][l] * TK[i][j-l];  
       TK[i][j] = (left->TK[i][j] - tmp2 - tmp) / right->TK[0][0];
     }
}


void operand::trigo2(i_function stf){

   int i,j,k,l;
   interval tmp, tmp2;

   if (!dep_t){ // stf(u(s))
     i = s_lb;
     if(!i){ tk[0] = stf(left->tk[0]); i++; }
     for(; i<= s_ub; i++){
        tmp = 0;
        for(k = 1; k < i; k++) 
          tmp += k * tk[k] * right->tk[i-k];             
        tk[i] = (left->tk[i] - tmp / i ) / right->tk[0];  
     }
     return;   
   }
   if (!dep_s){ // stf(u(t)
     j = t_lb;
     if(!j){ tk[0] = stf(left->tk[0]); j++; }
     for(; j<= t_ub; j++){
        tmp = 0;
        for(l = 1; l < j; l++) 
          tmp += l * tk[l] * right->tk[j-l];             
        tk[j] = - (left->tk[j] + tmp / j ) / right->tk[0];  
     }
     return;   
   }
       // stf(u(s,t))
   TK[0][0] = stf(left->TK[0][0]);
   for(j = 1; j <= t_ub; j++){
      tmp = 0;
      for(l = 1; l < j; l++) 
        tmp += l * TK[0][l] * right->TK[0][j-l];             
      TK[0][j] = - (left->TK[0][j] + tmp / j ) / right->TK[0][0];
   }
   for(i = 1; i<=s_ub; i++) 
     for(j = 0; j<=t_ub; j++){
       tmp2 = 0;
       for(k = 1; k < i; k++){
         tmp = 0;
         for(l = 0; l <= j; l++)
           tmp += right->TK[i-k][j-l] * TK[k][l]; 
           tmp2 += k * tmp;                          
       }
       tmp2 /= i;
       tmp = 0;
       for(l = 1; l <= j; l++)
          tmp += right->TK[0][l] * TK[i][j-l];  
       TK[i][j] = ( - left->TK[i][j] - tmp2 - tmp) / right->TK[0][0];
     }
}




// ----------------------------------------------------------------------------
// internal functions diff (implementation of taylorarithmetic)
// ----------------------------------------------------------------------------

void operand::diff(){

  if (status == done) return;

  interval *tmp_t, *tmp_s, **TMP, tmp, tmp2;
  int i, j, k, l;

  switch(o_type){

  case Const : break;

  case Uminus :  
       left->diff();
       if (!dep_s){ for(i = t_lb; i<= t_ub; i++) tk[i] = - left->tk[i]; break;}
       if (!dep_t){ for(i = s_lb; i<= s_ub; i++) tk[i] = - left->tk[i]; break;}
       for(i = 0; i<=s_ub; i++) 
          for(j = 0; j<=t_ub; j++) TK[i][j] = - left->TK[i][j];
       break;       
 
  case Plus : 
       left->diff(); right->diff(); 
       if (!dep_t){
         for(i = s_lb; i<= s_ub; i++) tk[i] = left->tk[i] + right->tk[i];
         break;   
       }
       if (!dep_s){
         for(i = t_lb; i<= t_ub; i++) tk[i] = left->tk[i] + right->tk[i];
         break;  
       }
       if ((!left->dep_s && !right->dep_t) || (!left->dep_t && !right->dep_s)){
	 if (left->dep_s){ tmp_s = left->tk; tmp_t = right->tk; } 
         else{ tmp_t = left->tk; tmp_s = right->tk;}
         TK[0][0] = tmp_s[0] + tmp_t[0];
         for(i = 1; i<= s_ub; i++) TK[i][0]  = tmp_s[i];        
         for(i = 1; i<= t_ub; i++) TK[0][i]  = tmp_t[i];
         for(i = 1; i<= s_ub; i++)
            for(j = 1; j<= t_ub; j++) TK[i][j] = 0;
         break;
       } 
       if (!left->dep_s || !right->dep_s){
         if(left->dep_s){ tmp_t = right->tk; TMP = left->TK;}
         else{ tmp_t = left->tk; TMP = right->TK;}
         for(j = 0; j<=t_ub; j++) TK[0][j] = TMP[0][j] + tmp_t[j];
         for(i = 1; i<=s_ub; i++)
            for( j = 0; j<=t_ub; j++) TK[i][j] = TMP[i][j];
         break;
       }
       if (!left->dep_t || !right->dep_t){
         if(left->dep_t){ tmp_s = right->tk; TMP = left->TK;}
         else{ tmp_s = left->tk; TMP = right->TK;}
         for(i = 0; i<=s_ub; i++){
            TK[i][0] = TMP[i][0] + tmp_s[i];       
            for(j = 1; j<=t_ub; j++) TK[i][j] = TMP[i][j];
	 }
         break;
       }
       for(i = 0; i<=s_ub; i++) 
         for(j = 0; j<=t_ub; j++) TK[i][j] = left->TK[i][j] + right->TK[i][j];
       break;
  
  case Minus :
       left->diff(); right->diff(); 
       if (!dep_t){
         for(i = s_lb; i<= s_ub; i++) tk[i] = left->tk[i] - right->tk[i];
         break;   
       }
       if (!dep_s){
         for(i = t_lb; i<= t_ub; i++) tk[i] = left->tk[i] - right->tk[i];
         break;  
       }
       if (!left->dep_t && !right->dep_s){
         TK[0][0] = left->tk[0] - right->tk[0];
         for(i = 1; i<= s_ub; i++) TK[i][0]  = left->tk[i];        
         for(i = 1; i<= t_ub; i++) TK[0][i]  = - right->tk[i];
         for(i = 1; i<= s_ub; i++)
           for(j = 1; j<= t_ub; j++) TK[i][j] = 0;
         break;
       }
       if (!left->dep_s && !right->dep_t){
         TK[0][0] = left->tk[0] - right->tk[0];
         for(i = 1; i<= s_ub; i++) TK[i][0]  = -right->tk[i];        
         for(i = 1; i<= t_ub; i++) TK[0][i]  = left->tk[i];
         for(i = 1; i<= s_ub; i++)
            for(j = 1; j<= t_ub; j++) TK[i][j] = 0;
         break;
       } 
       if (!left->dep_s){
	 // right->dep_s and right->dep_t
         for(j = 0; j<=t_ub; j++) TK[0][j] = left->tk[j] - right->TK[0][j];
         for(i = 1; i<=s_ub; i++)
            for( j = 0; j<=t_ub; j++) TK[i][j] = -right->TK[i][j];
         break; 
       }
       if( !right->dep_s){ 
	 // left->dep_s and left->dep_t
         for(j = 0; j<=t_ub; j++) TK[0][j] = left->TK[0][j] -  right->tk[j];
         for(i = 1; i<=s_ub; i++)
            for( j = 0; j<=t_ub; j++) TK[i][j] = left->TK[i][j];
         break;
       }
       if (!left->dep_t){
	 // right->dep_s and right->dep_t
         for(i = 0; i<=s_ub; i++){
           TK[i][0] = left->tk[i] -  right->TK[i][0];       
           for(j = 1; j<=t_ub; j++) TK[i][j] = -  right->TK[i][j];
	 }
         break;
       }
       if(!right->dep_t){
        //  left->dep_s and left->dep_t
         for(i = 0; i<=s_ub; i++){
           TK[i][0] = left->TK[i][0] - right->tk[i];       
           for(j = 1; j<=t_ub; j++) TK[i][j] = left->TK[i][j];
	 }
         break;
       }
       for(i = 0; i<=s_ub; i++) 
         for(j = 0; j<=t_ub; j++) TK[i][j] = left->TK[i][j] - right->TK[i][j];
       break;

 case Mult : 

       left->diff(); right->diff(); 
       if (!dep_t){
         for(i = s_lb; i<= s_ub; i++){
           i_accu = 0; 
           for(j = 0; j<=i; j++) accumulate(i_accu, left->tk[j],right->tk[i-j]);
           tk[i] = rnd(i_accu);
	 } 
         break;   
       }
       if (!dep_s){
         for(i = t_lb; i<= t_ub; i++){
           i_accu = 0; 
           tk[i] = 0;
           for(j = 0; j<=i; j++) accumulate(i_accu, left->tk[j], right->tk[i-j]);
           tk[i] = rnd(i_accu);
	 } 
         break;  
       }
       if ((!left->dep_s && !right->dep_t) || (!left->dep_t && !right->dep_s)){
	 if (left->dep_s){ tmp_s = left->tk; tmp_t = right->tk; } 
         else{ tmp_t = left->tk; tmp_s = right->tk;}
         for(i = 0; i<= s_ub; i++)
           for(j = 0; j<= t_ub; j++) TK[i][j]  = tmp_s[i] * tmp_t[j];
         break;
       } 
       if (!left->dep_s || !right->dep_s){
         if(left->dep_s){ tmp_t = right->tk; TMP = left->TK;}
         else{ tmp_t = left->tk; TMP = right->TK;}
         for(i = s_lb; i <= s_ub; i++)
           for(j = t_lb; j <= t_ub; j++){
             i_accu = 0;  
             for(k = 0; k<=j; k++) accumulate(i_accu, tmp_t[k], TMP[i][j-k]);
             TK[i][j] = rnd(i_accu);
	   }
         break;
       }
       if (!left->dep_t || !right->dep_t){
         if(left->dep_t){ tmp_s = right->tk; TMP = left->TK;}
         else{ tmp_s = left->tk; TMP = right->TK;}
         for(i = s_lb; i <= s_ub; i++)
           for(j = t_lb; j <= t_ub; j++){
             i_accu = 0;    
             for(k = 0; k<=i; k++) accumulate(i_accu, tmp_s[k], TMP[i-k][j]);
             TK[i][j] = rnd(i_accu); 
	   }
         break;
       }
       for(i = 0; i<=s_ub; i++) 
         for(j = 0; j<=t_ub; j++){
           i_accu = 0;
	   for(k = 0; k<=i; k++)
             for(l = 0; l <=j; l++)
               accumulate(i_accu, left->TK[k][l], right->TK[i-k][j-l]);
           TK[i][j] = rnd(i_accu); 
	 }
       break;


  case Div :
       left->diff(); right->diff(); 
       if (!dep_t){ // u(s) / v(s)
         for(i = s_lb; i<= s_ub; i++){
           i_accu = -left->tk[i];
           for(k = 1; k <= i; k++) accumulate(i_accu, right->tk[k], tk[i-k]);
           tk[i] = - rnd(i_accu) / right->tk[0]; 
	 }
         break;   
       }
       if (!dep_s){ // u(t) / v(t)
         for(j = t_lb; j<= t_ub; j++){
           i_accu= - left->tk[j];
           for(l = 1; l <= j; l++) accumulate(i_accu, right->tk[l], tk[j-l]);
           tk[j] = -rnd(i_accu) / right->tk[0]; 
	 }
         break;  
       }

       if (!left->dep_s && !right->dep_t){ // u(t) / v(s)
         for(j = 0; j <= t_ub; j++) TK[0][j] = left->tk[j] / right->tk[0];
         for(i = 1; i <= s_ub; i++)
           for(j = 0; j <= t_ub; j++) 
           {
             i_accu = 0;
             for(k = 1; k <= i; k++) accumulate(i_accu,right->tk[k], TK[i-k][j]);
             TK[i][j] = -rnd(i_accu) / right->tk[0];
	   }
         break;
       } 

       if (!left->dep_t && !right->dep_s){ // u(s) / v(t)
         for(i = 0; i <= s_ub; i++) TK[i][0] = left->tk[i] / right->tk[0];
         for(j = 1; j <= t_ub; j++)
           for(i = 0; i <= s_ub; i++) 
           {
             i_accu = 0;
             for(l = 1; l <= j; l++) accumulate(i_accu, right->tk[l],TK[i][j-l]);
             TK[i][j] = -rnd(i_accu) / right->tk[0];
	   }
         break;
       }
      
       if (!left->dep_s){ // u(t) / v(s,t)
         for(j = 0; j <= t_ub; j++){ 
          i_accu = -left->tk[j];
          for(l = 1; l <= j; l++) accumulate(i_accu, right->TK[0][l],TK[0][j-l]);
           TK[0][j] = -rnd(i_accu) / right->TK[0][0];
	 }
         for(i = 1; i <= s_ub; i++)
           for(j = 0; j <= t_ub; j++) 
           {
             i_accu = 0;
             for(k = 1; k <= i; k++)
               for(l = 0; l <= j; l++) 
                 accumulate(i_accu,right->TK[k][l],TK[i-k][j-l]);
             for(l = 1; l <= j; l++) 
               accumulate(i_accu, right->TK[0][l], TK[i][j-l]);
             TK[i][j] = -rnd(i_accu) / right->TK[0][0];
	   }
         break; 
       }

       if( !right->dep_s){  //  u(s,t) / v(t)
        for(i = 0; i <= s_ub; i++)
           for(j = 0; j <= t_ub; j++) 
           {
             i_accu = -left->TK[i][j];
             for(l = 1; l <= j; l++) accumulate(i_accu, right->tk[l],TK[i][j-l]);
             TK[i][j] = -rnd(i_accu) / right->tk[0];
	   }
         break;
       }

       if (!left->dep_t){ // u(s) / v(s,t)
         for(i = 0; i <= s_ub; i++){ 
           i_accu = -left->tk[i];
           for(k = 1; k <= i; k++) 
             accumulate(i_accu, right->TK[k][0], TK[i-k][0]);  
           TK[i][0] = -rnd(i_accu) / right->TK[0][0];
	 }
         for(i = 0; i <= s_ub; i++)
           for(j = 1; j <= t_ub; j++) 
           {
             i_accu = 0;
             for(k = 1; k <= i; k++)
               for(l = 0; l <= j; l++) 
                 accumulate(i_accu, right->TK[k][l], TK[i-k][j-l]);
             for(l = 1; l <= j; l++) 
               accumulate(i_accu, right->TK[0][l], TK[i][j-l]);
             TK[i][j] = -rnd(i_accu) / right->TK[0][0];
	   }
         break;
       }

       if(!right->dep_t){ //  u(s,t) / v(s)
         for(i = 0; i <= s_ub; i++)
           for(j = 0; j <= t_ub; j++) 
           {
             i_accu = -left->TK[i][j];
             for(k = 1; k <= i; k++) accumulate(i_accu,right->tk[k], TK[i-k][j]);
             TK[i][j] = -rnd(i_accu) / right->tk[0];
	   }
         break;
       }
       //  u(s,t) / v(s,t) 
       for(i = 0; i<=s_ub; i++) 
         for(j = 0; j<=t_ub; j++){
             i_accu = -left->TK[i][j];
             for(k = 1; k <= i; k++)
               for(l = 0; l <= j; l++) 
                 accumulate(i_accu, right->TK[k][l], TK[i-k][j-l]);
             for(l = 1; l <= j; l++) 
                 accumulate(i_accu, right->TK[0][l], TK[i][j-l]);
             TK[i][j] = -rnd(i_accu) / right->TK[0][0];
	 }
       break;

    case Exp :
       left->diff();
       if (!dep_t){ // exp(u(s))
         k = s_lb;
         if(!s_lb){ 
           tk[0] = exp(left->tk[0]); k++;} 
         for(i = k ; i<= s_ub; i++){
           tmp = 0;
           for(k = 0; k < i; k++) 
             tmp += ( i - k) * tk[k] * left->tk[i-k];
           tk[i] = tmp/ i; 
	 }
         break;   
       }
       if (!dep_s){ // exp(u(t))
         k = t_lb;
         if(!t_lb){ tk[0] = exp(left->tk[0]);
         k ++;} 
         for(j = k; j<= t_ub; j++){
           tmp = 0;
           for(l = 0; l < j; l++) 
             tmp += ( j - l) * tk[l] * left->tk[j-l];
           tk[j] = tmp / j; 
	 }
         break;  
       }
       //  exp(u(s,t))
       TK[0][0] = exp(left->TK[0][0]);
       for(i = 1 ; i<= s_ub; i++){
         tmp = 0;
         for(k = 0; k < i; k++){ 
           tmp += ( i - k) * TK[k][0] * left->TK[i-k][0];
           TK[i][0] = tmp/ i; 
	 }
       }
       for(i = 0; i<=s_ub; i++) 
         for(j = 1; j<=t_ub; j++){
           tmp = 0;
           for(l = 0; l < j; l++){ 
             i_accu = 0;
             for(k = 0; k <= i; k++) 
               accumulate(i_accu, TK[k][l], left->TK[i-k][j-l]);
             tmp += (j - l) * rnd(i_accu); 
	   }
           TK[i][j] = tmp / j;
	 }
       break;

    case Ln :
       left->diff(); 
       if (!dep_t){ // ln(u(s))
        i = s_lb;
        if(!i){ tk[0] = ln(left->tk[0]); i++; }
        for(; i<= s_ub; i++){
           tmp = 0;
           for(k = 1; k < i; k++) 
             tmp += k * tk[k] * left->tk[i-k];             
           tk[i] = (left->tk[i] - tmp / i ) / left->tk[0];  
	}
         break;   
       }
       if (!dep_s){ // ln(u(t)
         j = t_lb;
        if(!j){ tk[0] = ln(left->tk[0]); j++; }
        for(; j<= t_ub; j++){
           tmp = 0;
           for(l = 1; l < j; l++) 
             tmp += l * tk[l] * left->tk[j-l];             
           tk[j] = (left->tk[j] - tmp / j ) / left->tk[0];  
	}
         break;   
       }
       // ln(u(s,t))
        TK[0][0] = ln(left->TK[0][0]);
        for(j = 1; j <= t_ub; j++){
           tmp = 0;
           for(l = 1; l < j; l++) 
             tmp += l * TK[0][l] * left->TK[0][j-l];             
           TK[0][j] = (left->TK[0][j] - tmp / j ) / left->TK[0][0];
	}
        for(i = 1; i<=s_ub; i++) 
          for(j = 0; j<=t_ub; j++){
            tmp2 = 0;
            for(k = 1; k < i; k++){
              tmp = 0;
              for(l = 0; l <= j; l++)
                tmp += left->TK[i-k][j-l] * TK[k][l]; 
              tmp2 += k * tmp;                          
	    }
          tmp2 /= i;
          tmp = 0;
          for(l = 1; l <= j; l++)
             tmp += left->TK[0][l] * TK[i][j-l];  
          TK[i][j] = (left->TK[i][j] - tmp2 - tmp) / left->TK[0][0];
	  }
       break;
   
   case Sin :
       left->diff();
       right->status = done;
       if (!dep_t){ // sin(s)
         i = s_lb;
         if(!i){ tk[0] = sin(left->tk[0]);
                 right->tk[0] = cos(left->tk[0]);
                 i++;
         }
         for(; i<= s_ub; i++){
           tmp = 0;     // sin node
           for(k = 0; k < i; k++)
             tmp += (k + 1) * right->tk[i-k-1] * left->tk[k+1];
           tk[i] = tmp / i;
           tmp = 0;    // cos node 
           for(k = 0; k < i; k++)
             tmp += (k + 1) * tk[i-k-1] * left->tk[k+1];
           right->tk[i] = -tmp / i;
         }
         break;   
       }
       if (!dep_s){ // sin(t)
         j = t_lb;
         if(!j){ tk[0] = sin(left->tk[0]);
                 right->tk[0] = cos(left->tk[0]);  
                 j++; 
         }
         for(; j<= t_ub; j++){
           tmp = 0; // sin node
           for(l = 0; l < j; l++)
             tmp += (l + 1) * right->tk[j-l-1] * left->tk[l+1];
           tk[j] = tmp / j;
           tmp = 0; // cos node
           for(l = 0; l < j; l++)
             tmp += (l + 1) * tk[j-l-1] * left->tk[l+1];
           right->tk[j] = - tmp / j;
         }
         break;   
       }
       TK[0][0] = sin(left->TK[0][0]);
       right->TK[0][0] = cos(left->TK[0][0]);
       for(j = 1; j <= t_ub; j++){
         tmp = 0;  // sin node
         for(l = 0; l < j; l++)
           tmp += (l + 1) * right->TK[0][j-l-1] * left->TK[0][l+1];
         TK[0][j] = tmp / j;
         tmp = 0; // cos node
         for(l = 0; l < j; l++)
           tmp += (l + 1) * TK[0][j-l-1] * left->TK[0][l+1];
         right->TK[0][j] = - tmp / j;
       }
       for(i = 1; i<=s_ub; i++) 
          for(j = 0; j<=t_ub; j++){
            tmp2 = 0;  // sin node
            for(k = 0; k < i; k++){
              tmp = 0;   
              for(l = 0; l <= j; l++)
                tmp += right->TK[i-k-1][l] * left->TK[k+1][j-l]; 
            tmp2 += (k + 1) * tmp; 
            }
            TK[i][j] = tmp2 / i; 
            tmp2 = 0;  // cos node
            for(k = 0; k < i; k++){
              tmp = 0;   
              for(l = 0; l <= j; l++)
                tmp += TK[i-k-1][l] * left->TK[k+1][j-l]; 
            tmp2 += (k + 1) * tmp; 
            }
            right->TK[i][j] = - tmp2 / i; 
	  }
      break;

   case Cos :
       left->diff();
       right->status = done;
       if (!dep_t){ // cos(s)
         i = s_lb;
         if(!i){ tk[0] = cos(left->tk[0]);
                 right->tk[0] = sin(left->tk[0]);
                 i++;
         }
         for(; i<= s_ub; i++){
           tmp = 0;     // cos node
           for(k = 0; k < i; k++)
             tmp += (k + 1) * right->tk[i-k-1] * left->tk[k+1];
           tk[i] = -tmp / i;
           tmp = 0;    // sin node 
           for(k = 0; k < i; k++)
             tmp += (k + 1) * tk[i-k-1] * left->tk[k+1];
           right->tk[i] = tmp / i;
         }
         break;   
       }
       if (!dep_s){ // cos(t)
         j = t_lb;
         if(!j){ tk[0] = cos(left->tk[0]);
                 right->tk[0] = sin(left->tk[0]);  
                 j++; 
         }
         for(; j<= t_ub; j++){
           tmp = 0; // cos node
           for(l = 0; l < j; l++)
             tmp += (l + 1) * right->tk[j-l-1] * left->tk[l+1];
           tk[j] = -tmp / j;
           tmp = 0; // sin node
           for(l = 0; l < j; l++)
             tmp += (l + 1) * tk[j-l-1] * left->tk[l+1];
           right->tk[j] =  tmp / j;
         }
         break;   
       }
       TK[0][0] = cos(left->TK[0][0]);
       right->TK[0][0] = sin(left->TK[0][0]);
       for(j = 1; j <= t_ub; j++){
         tmp = 0;  // cos node
         for(l = 0; l < j; l++)
           tmp += (l + 1) * right->TK[0][j-l-1] * left->TK[0][l+1];
         TK[0][j] = - tmp / j;
         tmp = 0; // sin node
         for(l = 0; l < j; l++)
           tmp += (l + 1) * TK[0][j-l-1] * left->TK[0][l+1];
         right->TK[0][j] =  tmp / j;
       }
       for(i = 1; i<=s_ub; i++) 
          for(j = 0; j<=t_ub; j++){
            tmp2 = 0;  // cos node
            for(k = 0; k < i; k++){
              tmp = 0;   
              for(l = 0; l <= j; l++)
                tmp += right->TK[i-k-1][l] * left->TK[k+1][j-l]; 
            tmp2 += (k + 1) * tmp; 
            }
            TK[i][j] = -tmp2 / i; 
            tmp2 = 0;  // sin node
            for(k = 0; k < i; k++){
              tmp = 0;   
              for(l = 0; l <= j; l++)
                tmp += TK[i-k-1][l] * left->TK[k+1][j-l]; 
            tmp2 += (k + 1) * tmp; 
            }
            right->TK[i][j] =  tmp2 / i; 
	  }
      break;

   case Sinh :
       left->diff();
       right->status = done;
       if (!dep_t){ // sinh(s)
         i = s_lb;
         if(!i){ tk[0] = sinh(left->tk[0]);
                 right->tk[0] = cosh(left->tk[0]);
                 i++;
         }
         for(; i<= s_ub; i++){
           tmp = 0;     // sinh node
           for(k = 0; k < i; k++)
             tmp += (k + 1) * right->tk[i-k-1] * left->tk[k+1];
           tk[i] = tmp / i;
           tmp = 0;    // cosh node 
           for(k = 0; k < i; k++)
             tmp += (k + 1) * tk[i-k-1] * left->tk[k+1];
           right->tk[i] = tmp / i;
         }
         break;   
       }
       if (!dep_s){ // sinh(t)
         j = t_lb;
         if(!j){ tk[0] = sinh(left->tk[0]);
                 right->tk[0] = cosh(left->tk[0]);  
                 j++; 
         }
         for(; j<= t_ub; j++){
           tmp = 0; // sinh node
           for(l = 0; l < j; l++)
             tmp += (l + 1) * right->tk[j-l-1] * left->tk[l+1];
           tk[j] = tmp / j;
           tmp = 0; // cosh node
           for(l = 0; l < j; l++)
             tmp += (l + 1) * tk[j-l-1] * left->tk[l+1];
           right->tk[j] =  tmp / j;
         }
         break;   
       }
       TK[0][0] = sinh(left->TK[0][0]);
       right->TK[0][0] = cosh(left->TK[0][0]);
       for(j = 1; j <= t_ub; j++){
         tmp = 0;  // sinh node
         for(l = 0; l < j; l++)
           tmp += (l + 1) * right->TK[0][j-l-1] * left->TK[0][l+1];
         TK[0][j] = tmp / j;
         tmp = 0; // cosh node
         for(l = 0; l < j; l++)
           tmp += (l + 1) * TK[0][j-l-1] * left->TK[0][l+1];
         right->TK[0][j] =  tmp / j;
       }
       for(i = 1; i<=s_ub; i++) 
          for(j = 0; j<=t_ub; j++){
            tmp2 = 0;  // sinh node
            for(k = 0; k < i; k++){
              tmp = 0;   
              for(l = 0; l <= j; l++)
                tmp += right->TK[i-k-1][l] * left->TK[k+1][j-l]; 
            tmp2 += (k + 1) * tmp; 
            }
            TK[i][j] = tmp2 / i; 
            tmp2 = 0;  // cosh node
            for(k = 0; k < i; k++){
              tmp = 0;   
              for(l = 0; l <= j; l++)
                tmp += TK[i-k-1][l] * left->TK[k+1][j-l]; 
            tmp2 += (k + 1) * tmp; 
            }
            right->TK[i][j] = tmp2 / i; 
	  }
      break;

   case Cosh :
       left->diff();
       right->status = done;
       if (!dep_t){ // cosh(s)
         i = s_lb;
         if(!i){ tk[0] = cosh(left->tk[0]);
                 right->tk[0] = sinh(left->tk[0]);
                 i++;
         }
         for(; i<= s_ub; i++){
           tmp = 0;     // cosh node
           for(k = 0; k < i; k++)
             tmp += (k + 1) * right->tk[i-k-1] * left->tk[k+1];
           tk[i] = tmp / i;
           tmp = 0;    // sinh node 
           for(k = 0; k < i; k++)
             tmp += (k + 1) * tk[i-k-1] * left->tk[k+1];
           right->tk[i] = tmp / i;
         }
         break;   
       }
       if (!dep_s){ // cosh(t)
         j = t_lb;
         if(!j){ tk[0] = cosh(left->tk[0]);
                 right->tk[0] = sinh(left->tk[0]);  
                 j++; 
         }
         for(; j<= t_ub; j++){
           tmp = 0; // cosh node
           for(l = 0; l < j; l++)
             tmp += (l + 1) * right->tk[j-l-1] * left->tk[l+1];
           tk[j] = tmp / j;
           tmp = 0; // sinh node
           for(l = 0; l < j; l++)
             tmp += (l + 1) * tk[j-l-1] * left->tk[l+1];
           right->tk[j] =  tmp / j;
         }
         break;   
       }
       TK[0][0] = cosh(left->TK[0][0]);
       right->TK[0][0] = sinh(left->TK[0][0]);
       for(j = 1; j <= t_ub; j++){
         tmp = 0;  // cosh node
         for(l = 0; l < j; l++)
           tmp += (l + 1) * right->TK[0][j-l-1] * left->TK[0][l+1];
         TK[0][j] =  tmp / j;
         tmp = 0; // sinh node
         for(l = 0; l < j; l++)
           tmp += (l + 1) * TK[0][j-l-1] * left->TK[0][l+1];
         right->TK[0][j] =  tmp / j;
       }
       for(i = 1; i<=s_ub; i++) 
          for(j = 0; j<=t_ub; j++){
            tmp2 = 0;  // cosh node
            for(k = 0; k < i; k++){
              tmp = 0;   
              for(l = 0; l <= j; l++)
                tmp += right->TK[i-k-1][l] * left->TK[k+1][j-l]; 
            tmp2 += (k + 1) * tmp; 
            }
            TK[i][j] = tmp2 / i; 
            tmp2 = 0;  // sinh node
            for(k = 0; k < i; k++){
              tmp = 0;   
              for(l = 0; l <= j; l++)
                tmp += TK[i-k-1][l] * left->TK[k+1][j-l]; 
            tmp2 += (k + 1) * tmp; 
            }
            right->TK[i][j] =  tmp2 / i; 
	  }
      break;

  case Pow :        
       left->diff(); 
       right->status = done;
       if (!dep_t){ // pow(s)
         i = s_lb;
         if(!i){ tk[0] = pow(left->tk[0], right->value); i++; }
         for(; i<= s_ub; i++){
           tmp = 0;
           for(k = 0; k < i; k++)
             tmp += (right->value * (i - k) - k) * left->tk[i-k] * tk[k];
           tk[i] = tmp / (i * left->tk[0]);
         }
         break;   
       }
       if (!dep_s){ // pow(t)
         j = t_lb;
         if(!j){ tk[0] = pow(left->tk[0], right->value); j++; }
         for(; j<= t_ub; j++){
           tmp = 0;
           for(l = 0; l < j; l++)
             tmp += (right->value * (j - l) - l) * left->tk[j-l] * tk[l];
           tk[j] = tmp / (j * left->tk[0]);
         }
         break;   
       }
       TK[0][0] = pow(left->TK[0][0], right->value);
       for(j = 1; j <= t_ub; j++){
         tmp = 0;
         for(l = 0; l < j; l++)
           tmp += (right->value * (j - l) - l) * left->TK[0][j-l] * TK[0][l];
         TK[0][j] = tmp / (j * left->TK[0][0]);
       }
       for(i = 1; i<=s_ub; i++) 
          for(j = 0; j<=t_ub; j++){
            tmp2 = 0;
            for(k = 0; k < i; k++){
              tmp = 0;   
              for(l = 0; l <= j; l++)
                tmp += TK[k][l] * left->TK[i-k][j-l]; 
            tmp2 += (right->value * (i - k) - k) * tmp; 
            }
            tmp2 /= i;
            for(l = 1; l <= j; l++)
              tmp2 -= left->TK[0][l] * TK[i][j-l];
            TK[i][j] = tmp2 / left->TK[0][0] ; 
	  }
      break;


  case Power :
       left->diff(); 
       right->status = done;
       if (!dep_t){  // power(s)
         i = s_lb;
         if(!i){ tk[0] = power(left->tk[0], right->exponent); i++; }
         for(; i<= s_ub; i++){
           tmp = 0;
           for(k = 0; k < i; k++)
             tmp += (right->exponent * (i - k) - k) * left->tk[i-k] * tk[k];
           tk[i] = tmp / (i * left->tk[0]);
         }
         break;   
       }
       if (!dep_s){ // power(t)
         j = t_lb;
         if(!j){ tk[0] = power(left->tk[0], right->exponent); j++; }
         for(; j<= t_ub; j++){
           tmp = 0;
           for(l = 0; l < j; l++)
             tmp += (right->exponent * (j - l) - l) * left->tk[j-l] * tk[l];
           tk[j] = tmp / (j * left->tk[0]);
         }
         break;   
       }
       TK[0][0] = power(left->TK[0][0], right->exponent);
       for(j = 1; j <= t_ub; j++){
         tmp = 0;
         for(l = 0; l < j; l++)
          tmp += (right->exponent * (j - l) - l) * left->TK[0][j-l] * TK[0][l];
         TK[0][j] = tmp / (j * left->TK[0][0]);
       }
       for(i = 1; i<=s_ub; i++) 
          for(j = 0; j<=t_ub; j++){
            tmp2 = 0;
            for(k = 0; k < i; k++){
              tmp = 0;   
              for(l = 0; l <= j; l++)
                tmp += TK[k][l] * left->TK[i-k][j-l]; 
              tmp2 += (right->exponent * (i - k) - k) * tmp; 
            }
            tmp2 /= i;
            for(l = 1; l <= j; l++)
              tmp2 -= left->TK[0][l] * TK[i][j-l];
            TK[i][j] = tmp2 / left->TK[0][0] ; 
	  }
      break;

  case Sqr :

     left->diff(); 
     if (!dep_t){  // sqr(s)
       i = s_lb;
       if(!i){ tk[0] = sqr(left->tk[0]); i++;}
       for(; i<= s_ub; i++){
           i_accu = 0; 
           for(k = 0; k<= (i-1)/2 ; k++) 
              accumulate(i_accu, 2 * left->tk[k], left->tk[i-k]);
           if ( !(i % 2) ) accumulate(i_accu, left->tk[i/2], left->tk[i/2]);
           tk[i] = rnd(i_accu);
	 } 
         break;   
     }
     if (!dep_s){  // sqr(t)
       j = t_lb;
       if(!j){ tk[0] = sqr(left->tk[0]); j++;}
       for(; j<= t_ub; j++){
           i_accu = 0; 
           for(l = 0; l<= (j-1)/2 ; l++) 
              accumulate(i_accu, 2 * left->tk[l], left->tk[j-l]);
           if ( !(j % 2)) accumulate(i_accu, left->tk[j/2], left->tk[j/2]);
           tk[j] = rnd(i_accu);
	 } 
         break;   
     }
     TK[0][0] = sqr(left->TK[0][0]);
     for(j = 1; j <= t_ub; j++){
        i_accu = 0;
        for(l = 0; l<= (j-1)/2 ; l++)
           accumulate(i_accu, 2 * left->TK[0][l], left->TK[0][j-l]);
        if ( !(j % 2) ) accumulate(i_accu, left->TK[0][j/2], left->TK[0][j/2]);
	TK[0][j] = rnd(i_accu);
     }
     for(i = 1; i<= s_ub; i++) 
          for(j = 0; j<=t_ub; j++){
            i_accu = 0;
            for(k = 0; k <= (i-1)/2; k++){
              for(l = 0; l <= j ; l++)
                accumulate( i_accu, 2 * left->TK[k][l], left->TK[i-k][j-l]);
	    }
	    if ( !(i % 2) ){
              for(l = 0; l <= (j-1)/2; l++)
                 accumulate(i_accu, 2*left->TK[i/2][l], left->TK[i/2][j-l]);
              if( !(j % 2) )               
                    accumulate(i_accu, left->TK[i/2][j/2], left->TK[i/2][j/2]);
	    }
            TK[i][j] = rnd(i_accu);
	  }
     break;

  case Sqrt : 
  left->diff(); 
     if (!dep_t){  // sqrt(s)
       i = s_lb;
       if(!i){ tk[0] = sqrt(left->tk[0]); i++;}
       for(; i<= s_ub; i++){
           i_accu = 0; 
           for(k = 1; k<= (i-1)/2 ; k++) 
              accumulate(i_accu, 2 * tk[k], tk[i-k]);
           if ( !(i % 2) ) accumulate(i_accu, tk[i/2], tk[i/2]);
           tk[i] = (left->tk[i] - rnd(i_accu)) / ( 2 * tk[0]);
	 } 
         break;   
     }
     if (!dep_s){  // sqrt(t)
       j = t_lb;
       if(!j){ tk[0] = sqrt(left->tk[0]); j++;}
       for(; j<= t_ub; j++){
           i_accu = 0; 
           for(l = 1; l<= (j-1)/2 ; l++) 
              accumulate(i_accu, 2 * tk[l], tk[j-l]);
           if ( !(j % 2)) accumulate(i_accu, tk[j/2], tk[j/2]);
           tk[j] = (left->tk[j] - rnd(i_accu)) / ( 2 * tk[0]);
	 } 
         break;   
     }
     TK[0][0] = sqrt(left->TK[0][0]);
     for(j = 1; j <= t_ub; j++){
        i_accu = 0;
        for(l = 1; l<= (j-1)/2 ; l++)
           accumulate(i_accu, 2 * TK[0][l], TK[0][j-l]);
        if ( !(j % 2) ) accumulate(i_accu, TK[0][j/2], TK[0][j/2]);
	TK[0][j] = (left->TK[0][j] -rnd(i_accu)) / ( 2 * TK[0][0]);
     }
     for(i = 1; i<= s_ub; i++) 
          for(j = 0; j<=t_ub; j++){
            i_accu = 0;
            for(k = 1; k <= (i-1)/2; k++){
              for(l = 0; l <= j ; l++)
                accumulate( i_accu, 2 * TK[k][l], TK[i-k][j-l]);
	    }
	    if ( !(i % 2) ){
              for(l = 0; l <= (j-1)/2; l++)
                 accumulate(i_accu, 2*TK[i/2][l], TK[i/2][j-l]);
              if( !(j % 2) )               
                    accumulate(i_accu, TK[i/2][j/2], TK[i/2][j/2]);
	    }
            for(l = 1; l <= j ; l++)
                accumulate( i_accu, 2 * TK[0][l], TK[i][j-l]);
            TK[i][j] = (left->TK[i][j] - rnd(i_accu)) / ( 2 * TK[0][0]);
	  }
     break;

  case Tan : 
     left->diff(); right->left->left = left; right->diff();
     trigo1(tan);
     right->left->left = 0;
     break; 
  case Cot : 
     left->diff(); right->left->left = left; right->diff();
     trigo2(cot);
     right->left->left = 0;
     break; 
  case Tanh : 
     left->diff(); right->left->left = left; right->diff();
     trigo1(tanh);
     right->left->left = 0;
     break; 
  case Coth :
     left->diff(); right->left->left = left; right->diff();
     trigo2(coth);
     right->left->left = 0;
     break; 
  case Atan :
     left->diff(); right->right->left = left; right->diff();
     trigo1(atan);
     right->right->left = 0;
     break; 
  case Acot :
     left->diff(); right->right->left = left; right->diff();
     trigo2(acot);
     right->right->left = 0;
     break; 
  case Atanh :
     left->diff(); right->right->left = left; right->diff();
     trigo1(atanh);
     right->right->left = 0;
     break; 
  case Acoth :
     left->diff(); right->right->left = left; right->diff();
     trigo1(acoth);
     right->right->left = 0;
     break; 
  case Asin :
     left->diff(); right->left->right->left = left; right->diff();
     trigo1(asin);
     right->left->right->left = 0;
     break; 
  case Acos :
     left->diff(); right->left->right->left = left; right->diff();
     trigo2(acos);
     right->left->right->left = 0;
     break; 
  case Asinh :
     left->diff(); right->left->left->left = left; right->diff();
     trigo1(asinh);
     right->left->left->left = 0;
     break; 
  case Acosh :
     left->diff(); right->left->left->left = left; right->diff();
     trigo1(acosh);
     right->left->left->left = 0;
     break; 
  default    : throw error_foperand(Inc);
  }

  status = done;
}




// ----------------------------------------------------------------------------
// internal functions for memory management (operand)
// ----------------------------------------------------------------------------

void operand::alloc_rectangle(){
 TK = new interval*[dim_var1 + 1];
 for(int i=0; i <= dim_var1; i++) TK[i] = new interval[dim_var2 + 1];   
}


void operand::alloc_triangle(){
 TK = new interval*[dim_var1 + 1];
 for(int i=0; i <= dim_var1; i++) TK[i] = new interval[dim_var2 + 1 - i];     
}


void operand::alloc_taylor(){
 int tmp;
 if (status == done) return;
 switch (o_type){

 case Const :
            tmp = (dim_var1 > dim_var2)? dim_var1: dim_var2;
            tk = new interval[tmp + 1]; tk[0] = value; 
            for(int i=1; i<=tmp; i++) tk[i] = 0; break;

 case Plus : case Minus : case Mult : 
 case Div :  case Power : case Pow :
   
            left->alloc_taylor();
            right->alloc_taylor();
            if (!dep_s){ tk = new interval[dim_var2 + 1]; break;}
            if (!dep_t){ tk = new interval[dim_var1 + 1]; break;}
            if (triangular){ alloc_triangle(); break;}  
            alloc_rectangle(); break;

 case Sin :   case Cos :   case Sinh :  case Cosh : 
 case Tan :   case Cot :   case Tanh :  case Coth :
 case Asin :  case Acos :  case Atan :  case Acot :
 case Asinh : case Acosh : case Atanh : case Acoth :    
            if (left) left->alloc_taylor();
            if (right) right->alloc_taylor();
            if (!dep_s){ tk = new interval[dim_var2 + 1]; break;}
            if (!dep_t){ tk = new interval[dim_var1 + 1]; break;}
            if (triangular){ alloc_triangle(); right->alloc_triangle();break;} 
            alloc_rectangle();      

 default :  if (left) left->alloc_taylor();
            if (!dep_s){ 
               tk = new interval[dim_var2 + 1]; 
               break;
            }
            if (!dep_t){ 
               tk = new interval[dim_var1 + 1]; 
               break;}
            if (triangular){   
               alloc_triangle(); 
               break;
            }  
            alloc_rectangle(); break;          
 }
 status = done;
}



void operand::free_taylor(){
  // free memory for taylorcoefficients
   if (status == done) return;
   if (dep_s && dep_t){ 
     if (TK) for(int i=0; i <= dim_var1; i++) 
                 if (TK[i]){ delete[] TK[i];}
     delete[] TK; TK = 0;
   }
   else if (tk){ delete[] tk; tk = 0;}
   if(left)  left->free_taylor();
   if(right) right->free_taylor();
   status = done;
}  


void operand::delete_operand(){
  // free memory of the operand
 left = 0;
 right = 0;
 delete this;
}


void operand::save_named(){
  if (!del){ 
    //  cout << " detected " << o_type << endl; // remove!!
    if (o_type == Var) return;
    for(int i=0; i<index_named; i++) if( named[i] == this) return; 
    named[index_named++] = this;
    //  cout << " saved " << o_type << endl; // remove !!
  }
  if(left)  left->save_named();
  if(right) right->save_named();
}



operand* cp_operand(operand* op){

  int index = operand::index_named;
 
  if (!op) return 0;
  if (op->del) return new 
     operand(op, cp_operand(op->left), cp_operand(op->right));
  switch(op->o_type){
   case operand::Var_1: 
             return operand::var1; break;
   case operand::Var_2: 
             return operand::var2; break;
   case operand::Var:   
             if (op->dep_s) return operand::var1; 
             else           return operand::var2; break;
   default :   
            for(int i=0; i < index; i++)
             if (op == operand::named[i]) return operand::named[i + max_named];
             operand::named[index] = op;
             operand::named[index + max_named] = new operand(op, 0, 0);        
             if(index > max_named - 2) throw;    
             operand::index_named++; 
             operand::named[index + max_named]->left  = cp_operand(op->left);
             operand::named[index + max_named]->right = cp_operand(op->right);
             return operand::named[index + max_named];
  }
}




// ----------------------------------------------------------------------------
// assignment operator =
// ----------------------------------------------------------------------------


void  operand::operator=(operand& ri){
  o_type = ri.o_type;
  left   = ri.left;
  right  = ri.right;
  value  = ri.value;
  dep_s  = ri.dep_s;
  dep_t  = ri.dep_t;
  status = false;
  del    = false;
  tk     = 0;
  if(ri.del) ri.delete_operand();
}


// void  operand::operator=(const real& ri){
//   o_type = Const;
//   left   = 0;
//   right  = 0;
//   value  = ri;
//   dep_s  = false;
//   dep_t  = false;
//   status = false;
//   del    = false;
//   tk     = 0;
// }



//-----------------------------------------------------------------------------
// transpose operand
// ----------------------------------------------------------------------------


void operand::transpose(){
  bool tmp;
  if (status == done) return;
  switch (o_type){
  case Var  :  case Const:
             tmp = dep_t; dep_t = dep_s; dep_s = tmp; break;
  case Plus : case Minus : case Mult : case Div : case Power : case Pow :

             tmp = dep_t; dep_t = dep_s; dep_s = tmp;
             left->transpose();
             right->transpose(); break;
 
  case Sin :   case Cos :   case Sinh :  case Cosh : 
  case Tan :   case Cot :   case Tanh :  case Coth :
  case Asin :  case Acos :  case Atan :  case Acot :
  case Asinh : case Acosh : case Atanh : case Acoth :   
 
             tmp = dep_t; dep_t = dep_s; dep_s = tmp;
             if (left) left->transpose();
             if (right) right->transpose(); break;
  default :  
             tmp = dep_t; dep_t = dep_s; dep_s = tmp;
             if (left) left->transpose(); break;
  }
  status = done;
}



// ----------------------------------------------------------------------------
// operators -,+,-,*,/
// ----------------------------------------------------------------------------

operand& operator-(operand& Left){
   return  *(new operand(operand::Uminus, Left));  
}

operand& operator+(operand& Left, operand& Right){
   return  *(new operand(operand::Plus, Left, Right));  
}

operand& operator-(operand& Left, operand& Right){
   return *(new operand(operand::Minus, Left, Right));  
}

operand& operator*(operand& Left, operand& Right){
   return  *(new operand(operand::Mult, Left, Right));  
}

operand& operator/(operand& Left, operand& Right){
   return  *(new operand(operand::Div, Left, Right));  
}



// ----------------------------------------------------------------------------
// functions  exp, ln, sqr, sqrt, pow, power, sin, cos, tan, cot,
//            asin, acos, atan, acot, sinh, cosh, tanh, coth, 
//            asinh, acosh, atanh, acoth
// ----------------------------------------------------------------------------

operand& exp(operand& Left){
    return *(new operand(operand::Exp, Left));
}

operand& ln(operand& Left){
    return *(new operand(operand::Ln, Left));
}


operand& sqr(operand& Left){
    return *(new operand(operand::Sqr, Left));
}


operand& sqrt(operand& Left){
    return *(new operand(operand::Sqrt, Left));
}


operand& pow(operand& Left, const real& Right){
   operand* tmp = new operand(Right);
   return    *(new operand(operand::Pow, Left, *tmp));
}

operand& power(operand& Left, int Right){
   operand* tmp  = new operand(Right); 
   tmp->exponent = Right;
   return    *(new operand(operand::Power, Left, *tmp));
}

operand& sin(operand& Left){
   operand* costmp = new operand(operand::Cos, Left.dep_s, Left.dep_t);
   operand* sintmp = new operand(operand::Sin, Left, *costmp);
   return    *sintmp;
}

operand& cos(operand& Left){
   operand* sintmp = new operand(operand::Sin, Left.dep_s, Left.dep_t);
   operand* costmp = new operand(operand::Cos, Left, *sintmp);
   return    *costmp;
}


operand& tan(operand& Left){
   operand* sintmp = new operand(operand::Sin, Left.dep_s, Left.dep_t );
   operand* costmp = new operand(operand::Cos, Left.dep_s, Left.dep_t );
   costmp->right = sintmp;
   operand* sqrtmp = new operand(operand::Sqr, *costmp);
   operand* tantmp = new operand(operand::Tan, Left, *sqrtmp);
   return    *tantmp;
}


operand& cot(operand& Left){
   operand* costmp = new operand(operand::Cos, Left.dep_s, Left.dep_t );
   operand* sintmp = new operand(operand::Sin, Left.dep_s, Left.dep_t );
   sintmp->right = costmp;
   operand* sqrtmp = new operand(operand::Sqr, *sintmp  );
   operand* cottmp = new operand(operand::Cot, Left, *sqrtmp);
   return    *cottmp;
}


operand& sinh(operand& Left){
   operand* coshtmp = new operand(operand::Cosh, Left.dep_s, Left.dep_t);
   operand* sinhtmp = new operand(operand::Sinh, Left, *coshtmp);
   return    *sinhtmp;
}


operand& cosh(operand& Left){
   operand* sinhtmp = new operand(operand::Sinh, Left.dep_s, Left.dep_t);
   operand* coshtmp = new operand(operand::Cosh, Left, *sinhtmp);
   return    *coshtmp;
}

operand& tanh(operand& Left){
   operand* sinhtmp = new operand(operand::Sinh, Left.dep_s, Left.dep_t );
   operand* coshtmp = new operand(operand::Cosh, Left.dep_s, Left.dep_t );
   coshtmp->right = sinhtmp;
   operand* sqrtmp = new operand(operand::Sqr, *coshtmp);
   operand* tanhtmp = new operand(operand::Tanh, Left, *sqrtmp);
   return    *tanhtmp;
}

operand& coth(operand& Left){
   operand* coshtmp = new operand(operand::Cosh, Left.dep_s, Left.dep_t );
   operand* sinhtmp = new operand(operand::Sinh, Left.dep_s, Left.dep_t );
   sinhtmp->right = coshtmp;
   operand* sqrtmp = new operand(operand::Sqr, *sinhtmp  );
   operand* cothtmp = new operand(operand::Coth, Left, *sqrtmp);
   return    *cothtmp;
}


operand& asin(operand& Left){
   operand* sqrtmp  = new operand(operand::Sqr, Left.dep_s, Left.dep_t );
   operand* eins    = new operand(1);
   operand* mintmp  = new operand(operand::Minus,  *eins, *sqrtmp);
   operand* sqrttmp = new operand(operand::Sqrt, *mintmp );
   operand* asintmp = new operand(operand::Asin, Left, *sqrttmp);
   return    *asintmp;
}

operand& acos(operand& Left){
   operand* sqrtmp  = new operand(operand::Sqr, Left.dep_s, Left.dep_t );
   operand* eins    = new operand(1);
   operand* mintmp  = new operand(operand::Minus, *eins, *sqrtmp);
   operand* sqrttmp = new operand(operand::Sqrt, *mintmp );
   operand* acostmp = new operand(operand::Acos, Left, *sqrttmp);
   return    *acostmp;
}

operand& atan(operand& Left){
   operand* sqrtmp  = new operand(operand::Sqr, Left.dep_s, Left.dep_t ); 
   operand* eins    = new operand(1);
   operand* plustmp = new operand(operand::Plus, *eins, *sqrtmp);
   operand* atantmp = new operand(operand::Atan, Left, *plustmp);
   return    *atantmp;
}

operand& acot(operand& Left){
   operand* sqrtmp  = new operand(operand::Sqr, Left.dep_s, Left.dep_t );
   operand* eins    = new operand(1); 
   operand* plustmp = new operand(operand::Plus, *eins, *sqrtmp);
   operand* acottmp = new operand(operand::Acot, Left, *plustmp);
   return    *acottmp;
}

operand& asinh(operand& Left){
   operand* sqrtmp  = new operand(operand::Sqr, Left.dep_s, Left.dep_t );
   operand* eins    = new operand(1);
   operand* plustmp = new operand(operand::Plus,  *sqrtmp, *eins);
   operand* sqrttmp = new operand(operand::Sqrt, *plustmp );
   operand* asinhtmp = new operand(operand::Asinh, Left, *sqrttmp);
   return    *asinhtmp;
}

operand& acosh(operand& Left){
   operand* sqrtmp  = new operand(operand::Sqr, Left.dep_s, Left.dep_t );
   operand* eins    = new operand(1);
   operand* mintmp  = new operand(operand::Minus, *sqrtmp, *eins);
   operand* sqrttmp = new operand(operand::Sqrt, *mintmp );
   operand* acoshtmp = new operand(operand::Acosh, Left, *sqrttmp);
   return    *acoshtmp;
}

operand& atanh(operand& Left){
   operand* sqrtmp   = new operand(operand::Sqr, Left.dep_s, Left.dep_t );
   operand* eins     = new operand(1);
   operand* mintmp   = new operand(operand::Minus, *eins, *sqrtmp);
   operand* atanhtmp = new operand(operand::Atanh, Left, *mintmp);
   return    *atanhtmp;
}

operand& acoth(operand& Left){
   operand* sqrtmp   = new operand(operand::Sqr, Left.dep_s, Left.dep_t ); 
   operand* eins    = new operand(1);
   operand* mintmp   = new operand(operand::Minus, *eins, *sqrtmp);
   operand* acothtmp = new operand(operand::Acoth, Left, *mintmp);
   return    *acothtmp;
}


string o_type2string(operand::operand_type op_type){

 string name;

 switch(op_type){
  case operand::Const : name = "const"; break; 
  case operand::Var_1 : name = "s!"; break;
  case operand::Var_2 : name = "t!"; break;
  case operand::Plus  : name = "+"; break; 
  case operand::Minus : name = "-"; break;
  case operand::Mult  : name = "*"; break;
  case operand::Div   : name = "/"; break;
  case operand::Exp   : name = "exp"; break;
  case operand::Ln    : name = "ln"; break;
  case operand::Sqr   : name = "sqr"; break;
  case operand::Sqrt  : name = "sqrt"; break;
  case operand::Sin   : name = "sin"; break; 
  case operand::Cos   : name = "cos"; break;
  case operand::Tan   : name = "tan"; break;
  case operand::Cot   : name = "cot"; break;
  case operand::Asin  : name = "asin"; break;
  case operand::Acos  : name = "acos"; break;
  case operand::Atan  : name = "atan"; break;
  case operand::Acot  : name = "acot"; break;
  case operand::Sinh  : name = "sinh"; break;
  case operand::Cosh  : name = "cosh"; break;
  case operand::Tanh  : name = "tanh"; break;
  case operand::Coth  : name = "coth"; break;
  case operand::Asinh : name = "asinh"; break;
  case operand::Acosh : name = "acosh"; break;
  case operand::Atanh : name = "atanh"; break;
  case operand::Acoth : name = "acoth"; break;
  case operand::Pow   : name = "pow"; break;  
  case operand::Power : name = "power"; break;
 }
 return name;
}


// ---------------------------------------------------------------------------
// ostream operators
// ---------------------------------------------------------------------------


ostream& operator<<(ostream& os, const operand& out){

    
    switch(out.o_type){
   
      case operand::Var_1: case operand::Var_2:

                    if (out.left || out.right) 
                       throw operand::error_foperand(operand::Var);
                    os << o_type2string(out.o_type);
                    if (out.detail) os << "[" << out.dep_s << out.dep_t 
                                       << out.status << "][" << &out << "]";  
                    break;
    
      case operand::Var:

                    if (out.left || out.right)
                       throw operand::error_foperand(out.o_type);
                    if (out.dep_t) os << "t";
                    else os << "s";
                    if (out.detail) os << "[" << out.dep_s << out.dep_t 
                                      << out.status << "][" << &out << "]";  
                    break;

      case operand::Const:

                    if (out.left || out.right) 
                       throw operand::error_foperand(out.o_type);
	            os  << " " << out.value<< " "; 
                    if (out.detail) os << "[" << out.dep_s << out.dep_t 
                                       << out.status << "][" << &out << "]";  
                    break;

      case operand::Uminus:

                    if (!out.left || out.right) 
                       throw operand::error_foperand(out.o_type);
                    os  << " -" <<  *out.left;
                    if (out.detail) os << "[" << out.dep_s << out.dep_t 
                                       << out.status << "][" << &out << "]";  
                    break;

      case operand::Plus:  case operand::Minus: 
      case operand::Mult:  case operand::Div:

                    if (!out.left || !out.right) 
                       throw operand::error_foperand(out.o_type);
		    os << " (" << *out.left  << o_type2string( out.o_type) 
                       << *out.right << ")"; 
                    if (out.detail) os << "[" << out.dep_s << out.dep_t 
				       << out.status << "][" << &out << "]";  
                    break;

 
      case operand::Exp: case operand::Ln: 
      case operand::Sqr: case operand::Sqrt:

                    if (!out.left || out.right)
                       throw operand::error_foperand(out.o_type);
	            os << o_type2string(out.o_type) <<"(" << *out.left << ")";
                    if (out.detail) os << "[" << out.dep_s << out.dep_t 
                                       << out.status << "][" << &out << "]";  
                    break;

  
      case operand::Sin:  case operand::Cos:
      case operand::Tan:  case operand::Asin:
      case operand::Acos: case operand::Atan:
      case operand::Sinh: case operand::Cosh:
      case operand::Tanh: case operand::Asinh:
      case operand::Acosh: case operand::Atanh:

                    if (!out.left || !out.right)
                        throw operand::error_foperand(out.o_type); 
                    os << o_type2string(out.o_type) << "(" << *out.left << ")";
                    if (out.detail) os << "[" << out.dep_s << out.dep_t 
                                       << out.status << "][" << &out << "]";  
                    break;

      case operand::Pow: case operand::Power:

                    if (!out.left || !out.right || 
                        out.right->o_type != operand::Const)
                        throw operand::error_foperand(out.o_type); 
                    os  <<  o_type2string(out.o_type) << "(" <<  *out.left 
			<< ","     << *out.right << ")";
                    if (out.detail) os << "[" << out.dep_s << out.dep_t 
                                       << out.status << "][" << &out << "]";  
                    break;

 

      default:      throw operand::error_foperand(operand::Inc);        
      }

  return os;
} 


// ============================================================================
// ----------------------------------------------------------------------------
// class integrand
// ----------------------------------------------------------------------------
// ============================================================================


// ----------------------------------------------------------------------------
// constructor and destructor
// ----------------------------------------------------------------------------

integrand::integrand(int dim1, int dim2){

  dim_var1 = dim1;
  dim_var2 = dim2;
  var1     = 0;
  var2     = 0;
  dep_s    = false;
  dep_t    = false;
  del      = false;
  diff_s   = -1;
  diff_t   = -1;
  ftree    = 0;
  status   = false;
}


integrand::integrand(const integrand& igr){

  dim_var1 = igr.dim_var1;
  dim_var2 = igr.dim_var2;
  var1     = 0;
  var2     = 0;
  set_var();
  dep_s    = igr.dep_s;
  dep_t    = igr.dep_t;
  del      = false;
  diff_s   = -1;
  diff_t   = -1;
  operand::index_named = 0;
  ftree    = cp_operand(igr.ftree);
  status   = false;
  if (igr.del) delete &igr;
}




void integrand::delete_ftree(){
  if(!ftree) return;
  operand::index_named = 0;
  ftree->save_named();
  if (ftree->del) delete ftree;
  for(int i=0; i<operand::index_named; i++) delete operand::named[i];
}


integrand::~integrand(){
  free_taylor();
  delete_ftree();
  if (var1){ delete var1; var1 = 0;}
  if (var2){ delete var2; var2 = 0;}
}


// ----------------------------------------------------------------------------
// functions alloc, free for memory management
// ----------------------------------------------------------------------------

void integrand::alloc_taylor(int dim1, int dim2){
  if (dim1 + dim2){
   operand::dim_var1 = dim_var1 = dim1;
   operand::dim_var2 = dim_var2 = dim2;
  }
  else{
   operand::dim_var1 = dim_var1;
   operand::dim_var2 = dim_var2;
  }
  status = operand::done = !status;
  if (var1){
    var1->tk = new interval[dim_var1 + 1 ];
    var1->status = status;
    for(int i = 0; i <= dim_var1; i++) var1->tk[i] = 0;
  }
  if (var2){
     var2->tk = new interval[dim_var2 + 1];
     var2->status = status;
     for(int i = 0; i <= dim_var2; i++) var2->tk[i] = 0;
  }

  if(ftree) ftree->alloc_taylor();
}


void integrand::free_taylor(){
  status = operand::done = !status;
  operand::dim_var1 = dim_var1;
  operand::dim_var2 = dim_var2;
  if(ftree) ftree->free_taylor();
}


// ----------------------------------------------------------------------------
// internal functions
//-----------------------------------------------------------------------------

inline void integrand::set_var1(const interval& s_value ,bool var) const{
  var1->tk[0] = s_value;
  if (!dim_var1) return; // no memory for tk[1] allocated
  if (var) var1->tk[1] = 1;
  else var1->tk[1] = 0;
}


inline void integrand::set_var2(const interval& t_value, bool var) const{
  var2->tk[0] = t_value;
  if (!dim_var2) return; // no memory for tk[1] allocated
  if (var) var2->tk[1] = 1;
  else var2->tk[1] = 0;
}


inline void integrand::set_var(){
  if (!var1){
      var1 = new operand(operand::Var_1);
      var1->o_type = operand::Var;
  }
  operand::var1 = var1;
      
  if(!var2){ 
      var2 = new operand(operand::Var_2);
      var2->o_type = operand::Var;
  }
  operand::var2 = var2;
}



inline void integrand::set_dim(int dim_s, int dim_t){
 free_taylor();
 dim_var1 = dim_s;
 dim_var2 = dim_t;
 alloc_taylor();
}


// ----------------------------------------------------------------------------
// public member function value
// ----------------------------------------------------------------------------


interval integrand::value(const interval& s){
  status = !status;
  if (var1){ var1->value = s; var1->status = status;}
  operand::done = status;
  ftree->set_value();
  return ftree->value;
}


interval integrand::value(const interval& s, const interval& t) const{
  status = !status;
  if (var1){ var1->value = s; var1->status = status;}
  if (var2){ var2->value = t; var2->status = status;}
  operand::done = status;
  ftree->set_value();
  return ftree->value;
}



// ----------------------------------------------------------------------------
// public member functions  taylor
// ----------------------------------------------------------------------------


interval integrand::taylor(int order_s) const {
  // starts computation of taylorcoeff. with unchanged argument
  if (dim_var1 < order_s)  throw error(1);
  if (diff_s   < order_s){
    operand::done = status = !status;
    if (var1) var1->status = status;
    if (var2) var2->status = status;
    operand::s_lb = diff_s + 1;
    operand::s_ub = order_s;
    operand::t_lb = 0;
    operand::t_ub = 0;
    ftree->diff();     // calculating tcoeffs. from diff_s + 1 to order_s
    diff_s = order_s;
    diff_t = 0;
  }
  if  (!dep_t) return ftree->tk[order_s];
  if  (!dep_s) if(order_s) return _interval(0);
               else return ftree->tk[0];
  else return ftree->TK[order_s][0];
}


interval integrand::taylor(int order_s, const interval& arg_s, 
                                        const interval& arg_t) const
{
  // starts computation of taylorcoeff. with argument arg_s
  if (dim_var1 < order_s)  throw error(1);
  operand::done = status = !status;
  if (var1) var1->status = status;
  if (var2) var2->status = status;
  if(dep_s) set_var1(arg_s);
  if(dep_t) set_var2(arg_t);
  operand::s_lb = 0;
  operand::s_ub = order_s;
  operand::t_lb = 0;
  operand::t_ub = 0;
  //  operand::arg_s= arg_s;
  //  operand::arg_t= arg_t;
  ftree->diff();    // berechnen von 0 bis order_s 
  diff_s = order_s;
  diff_t = 0;     
  if  (!dep_t) return ftree->tk[order_s];
  if  (!dep_s) if(order_s) return _interval(0); 
               else return ftree->tk[0];
  else return ftree->TK[order_s][0];
}


interval integrand::taylor(int order_s, int order_t) const
{
  if (dim_var1 < order_s || dim_var2 < order_t) throw error(1);     
  if (diff_s   < order_s || diff_t < order_t){
    operand::done = status = !status;
    var1->status = status;
    var2->status = status;
    operand::s_lb = 0;
    operand::s_ub = order_s;
    operand::t_lb = 0;
    operand::t_ub = order_t;
    ftree->diff();     
    diff_s = order_s;
    diff_t = order_t;    
  }
  if  (!dep_t) if (order_t) return _interval(0);
               else return ftree->tk[order_s];
  if  (!dep_s) if (order_s) return _interval(0);
               else return ftree->tk[order_t];
  return ftree->TK[order_s][order_t];
}


interval integrand::taylor(int order_s, int order_t, const interval& arg_s,
                                                  const interval& arg_t) const
{
  if (dim_var1 < order_s || dim_var2 < order_t) throw error(1);  
  operand::done = status = !status;
  var1->status = status;
  var2->status = status;
  if (dep_s) set_var1(arg_s);
  if (dep_t) set_var2(arg_t); 
  operand::s_lb = 0;
  operand::s_ub = order_s;
  operand::t_lb = 0;
  operand::t_ub = order_t;
  ftree->diff();
  diff_s = order_s;
  diff_t = order_t;  
  if  (!dep_t) if (order_t) return _interval(0);
               else return ftree->tk[order_s];
  if  (!dep_s) if (order_s) return _interval(0);
               else return ftree->tk[order_t];
  return ftree->TK[order_s][order_t];
}



// ----------------------------------------------------------------------------
// integrand operator =, operator<< 
// ----------------------------------------------------------------------------

integrand& integrand::operator=(operand& right){

  // free memory left side
  if (ftree) free_taylor();
  delete_ftree();
  // allocate memory for variables
  set_var();
  dep_s = right.dep_s;
  dep_t = right.dep_t;

  // special case: right side is a variable
  if (right.o_type == operand::Var_1){
    ftree = var1;
    alloc_taylor();
    return *this;
  }
 
  if (right.o_type == operand::Var_2){
    ftree = var2;
    alloc_taylor();
    return *this;
  }

  // generate deep copy
  operand::index_named = 0;
  ftree  = cp_operand(&right);
  // delete right side if necessary
  if(right.del) delete &right;
  status = false;
  // allocate memory for taylorcoeffs.
  free_taylor();
  alloc_taylor(); 
  return *this;
} 


integrand& integrand::operator=(integrand& right){

  if (this == &right) return *this;
  free_taylor();
  if (ftree) delete_ftree();
  dep_s    = right.dep_s;
  dep_t    = right.dep_t;

  if (right.del){ 
    status   = right.status;
    if(var1) var1->delete_operand();
    var1     = right.var1;
    if(var2) var2->delete_operand();
    var2     = right.var2;
    ftree    = right.ftree;
    free_taylor();
    right.ftree = 0; right.var1 = 0; right.var2 = 0;
    delete &right;
  }

  else{  // right.del == false
    status = false;
    set_var();
    operand::index_named = 0;
    ftree = cp_operand(right.ftree);
  }
  alloc_taylor();
  return *this;
}


ostream& operator<<(ostream& os, const integrand& out){
  if(out.ftree)  os << *out.ftree;
  return os;
}



// ----------------------------------------------------------------------------
// public operator integrand x integrand  x in { +,-,*,/ }
// ----------------------------------------------------------------------------


integrand& integrand::operator+(integrand& right){

  integrand* result = new integrand;
  operand*   tmp;

  result->del = true;
  result->set_var();
  result->dep_s = dep_s || right.dep_s ;
  result->dep_t = dep_t || right.dep_t ;
  tmp = new operand(operand::Plus, *this->ftree, *right.ftree); 
  operand::index_named = 0;
  result->ftree = cp_operand(tmp);
  tmp->delete_operand();
  if(right.del) delete &right;
  if(this->del) delete this;  
  return *result;
}



integrand& integrand::operator-(integrand& right){

  integrand* result = new integrand;
  operand*   tmp;

  result->del = true;
  result->set_var();
  result->dep_s = dep_s || right.dep_s ;
  result->dep_t = dep_t || right.dep_t ;
  tmp = new operand(operand::Minus, *this->ftree, *right.ftree); 
  operand::index_named = 0;
  result->ftree = cp_operand(tmp);
  tmp->delete_operand();
  if(right.del) delete &right;
  if(this->del) delete this;  
  return *result;
}



integrand& integrand::operator*(integrand& right){

  integrand* result = new integrand;
  operand*   tmp;

  result->del = true;
  result->set_var();
  result->dep_s = (dep_s || right.dep_s) ;
  result->dep_t = (dep_t || right.dep_t) ;
  tmp = new operand(operand::Mult, *this->ftree, *right.ftree); 
  operand::index_named = 0;
  result->ftree = cp_operand(tmp);
  tmp->delete_operand();
  if(right.del) delete &right;
  if(this->del) delete this;  
  return *result;
}



integrand& integrand::operator/(integrand& right){

  integrand* result = new integrand;
  operand*   tmp;

  result->del = true;
  result->set_var();
  result->dep_s = (dep_s || right.dep_s) ;
  result->dep_t = (dep_t || right.dep_t) ;
  tmp = new operand(operand::Div, *this->ftree, *right.ftree); 
  operand::index_named = 0;
  result->ftree = cp_operand(tmp);
  tmp->delete_operand();
  if(right.del) delete &right;
  if(this->del) delete this;  
  return *result;
}



// ----------------------------------------------------------------------------
// public operator  integrand x operator, member function transpose
// ----------------------------------------------------------------------------


integrand& integrand::operator+(operand& right){
  integrand* result;
  if(del){
    result = new integrand(dim_var1, dim_var2);
    result->var1 = operand::var1 = var1;
    result->var2 = operand::var2 = var2;
    result->ftree = ftree;
    result->dep_s = dep_s;
    result->dep_t = dep_t;
    ftree = 0; var1 = 0; var2 = 0; delete this; 
  }
  else
    result = new integrand(*this); // copy constructor call
  result->del = true;
  operand::index_named = 0;
  result->ftree = new 
     operand(operand::Plus, *result->ftree, *cp_operand(&right));  
  if (right.del) delete &right;
  return *result;
}


integrand& integrand::operator-(operand& right){
  integrand* result;
  if(del){
    result = new integrand(dim_var1, dim_var2);
    result->var1 = operand::var1 = var1;
    result->var2 = operand::var2 = var2;
    result->ftree = ftree;
    result->dep_s = dep_s;
    result->dep_t = dep_t;
    ftree = 0; var1 = 0; var2 = 0; delete this; 
  }
  else
    result = new integrand(*this); // copy constructor call
  result->del = true;
  operand::index_named = 0;
  result->ftree = new 
     operand(operand::Minus, *result->ftree, *cp_operand(&right));  
  if (right.del) delete &right;
  return *result;
}


integrand& integrand::operator*(operand& right){
  integrand* result;
  if(del){
    result = new integrand(dim_var1, dim_var2);
    result->var1 = operand::var1 = var1;
    result->var2 = operand::var2 = var2;
    result->ftree = ftree;
    result->dep_s = dep_s;
    result->dep_t = dep_t;
    ftree = 0; var1 = 0; var2 = 0; delete this; 
  }
  else
    result = new integrand(*this); // copy constructor call
  result->del = true;
  operand::index_named = 0;
  result->ftree = new 
     operand(operand::Mult, *result->ftree, *cp_operand(&right));  
  if (right.del) delete &right;
  return *result;
}



integrand& integrand::operator/(operand& right){
  integrand* result;
  if(del){
    result = new integrand(dim_var1, dim_var2);
    result->var1 = operand::var1 = var1;
    result->var2 = operand::var2 = var2;
    result->ftree = ftree;
    result->dep_s = dep_s;
    result->dep_t = dep_t;
    ftree = 0; var1 = 0; var2 = 0; delete this; 
  }
  else
    result = new integrand(*this); // copy constructor call
  result->del = true;
  operand::index_named = 0;
  result->ftree = new 
     operand(operand::Div, *result->ftree, *cp_operand(&right));  
  if (right.del) delete &right;
  return *result;
}


integrand& integrand::transpose(){
  integrand* result = new integrand(*this);  
  operand* tmp;
  tmp          = result->var1;
  result->var1 = result->var2;
  result->var2 = tmp;
  result->del  = true; 
  result->status = operand::done = !result->status;
  result->ftree->transpose();
  result->dep_s = dep_t;
  result->dep_t = dep_s;
  if (del) delete this;
  return *result;
}

